import json
import os

from .dict import ToolConfDict


class ToolConfJsonFile(ToolConfDict):
    _configs_dir = None

    def __init__(self, config_file):
        # type: (str) -> None
        """
        config_file contains JSON with issuers settings.
        Each key is issuer and value is issuer's configuration.
        Configuration could be set in two formats:

        1. { ... "iss": { ... "client_id: "client" ... }, ... }
        In this case the library will work in the concept: one issuer ~ one client-id

        2. { ... "iss": [ { ... "client_id: "client1" ... }, { ... "client_id: "client2" ... } ], ... }
        In this case the library will work in concept: one issuer ~ many client-ids

        Example:
            {
                "iss1": [{
                        "default": true,
                        "client_id": "client_id1",
                        "auth_login_url": "auth_login_url1",
                        "auth_token_url": "auth_token_url1",
                        "auth_audience": null,
                        "key_set_url": "key_set_url1",
                        "key_set": null,
                        "private_key_file": "private.key",
                        "public_key_file": "public.key",
                        "deployment_ids": ["deployment_id1", "deployment_id2"]
                    }, {
                        "default": false,
                        "client_id": "client_id2",
                        "auth_login_url": "auth_login_url2",
                        "auth_token_url": "auth_token_url2",
                        "auth_audience": null,
                        "key_set_url": "key_set_url2",
                        "key_set": null,
                        "private_key_file": "private.key",
                        "public_key_file": "public.key",
                        "deployment_ids": ["deployment_id3", "deployment_id4"]
                    }],
                "iss2": [ .... ]
            }

        default (bool) - this iss config will be used in case if client-id was not passed on the login step
        client_id - this is the id received in the 'aud' during a launch
        auth_login_url - the platform's OIDC login endpoint
        auth_token_url - the platform's service authorization endpoint
        auth_audience - the platform's OAuth2 Audience (aud). Is used to get platform's access token,
                        Usually the same as "auth_token_url" but in the common case could be a different url
        key_set_url - the platform's JWKS endpoint
        key_set - in case if platform's JWKS endpoint somehow unavailable you may paste JWKS here
        private_key_file - relative path to the tool's private key
        public_key_file - relative path to the tool's public key
        deployment_ids (list) - The deployment_id passed by the platform during launch
        """
        if not os.path.isfile(config_file):
            raise Exception("LTI tool config file not found: " + config_file)
        self._configs_dir = os.path.dirname(config_file)

        cfg = open(config_file, 'r')
        iss_conf_dict = json.loads(cfg.read())
        super(ToolConfJsonFile, self).__init__(iss_conf_dict)
        cfg.close()

        for iss in iss_conf_dict:
            if isinstance(iss_conf_dict[iss], list):
                for iss_conf in iss_conf_dict[iss]:
                    self._process_iss_conf_item(iss_conf, iss, iss_conf['client_id'])
            else:
                self._process_iss_conf_item(iss_conf_dict[iss], iss)

    def _process_iss_conf_item(self, iss_conf, iss, client_id=None):
        private_key_file = iss_conf['private_key_file']
        if not private_key_file.startswith('/'):
            private_key_file = self._configs_dir + '/' + private_key_file

        prf = open(private_key_file, 'r')
        self.set_private_key(iss, prf.read(), client_id=client_id)
        prf.close()

        public_key_file = iss_conf.get('public_key_file', None)
        if public_key_file:
            if not public_key_file.startswith('/'):
                public_key_file = self._configs_dir + '/' + public_key_file

            pubf = open(public_key_file, 'r')
            self.set_public_key(iss, pubf.read(), client_id=client_id)
            pubf.close()



import typing as t
from abc import ABCMeta, abstractmethod

if t.TYPE_CHECKING:
    from typing_extensions import Protocol

    class SessionLike(Protocol):
        @t.overload
        def get(self, key):
            # type: (str) -> t.Any
            pass

        @t.overload
        def get(self, key, default):  # pylint: disable=function-redefined
            # type: (str, t.Any) -> t.Any
            pass

        def get(self, key, default=None):  # pylint: disable=function-redefined
            # type: (str, t.Any) -> t.Any
            pass

        def __setitem__(self, key, value):
            # type: (str, t.Any) -> None
            pass

        def __contains__(self, key):
            # type: (object) -> bool
            pass


class Request(object):
    __metaclass__ = ABCMeta

    @property
    def session(self):
        # type: () -> SessionLike
        raise NotImplementedError

    @abstractmethod
    def is_secure(self):
        # type: () -> bool
        raise NotImplementedError

    @abstractmethod
    def get_param(self, key):
        # type: (str) -> object
        raise NotImplementedError



import typing as t
from abc import ABCMeta, abstractmethod

if t.TYPE_CHECKING:
    from ..request import Request

_T_DISABLED_SESSION_ID = t.TypeVar('_T_DISABLED_SESSION_ID', bound='DisableSessionId')
T = t.TypeVar('T')


class LaunchDataStorage(t.Generic[T]):
    __metaclass__ = ABCMeta
    _request = None  # type: t.Optional[Request]
    _session_id = None  # type: t.Optional[str]
    _session_cookie_name = 'session-id'  # type: str
    _prefix = 'lti1p3-'  # type: str

    def __init__(self, *args, **kwargs):
        # type: (*t.Any, **t.Any) -> None
        pass

    def set_request(self, request):
        # type: (Request) -> None
        self._request = request

    def get_session_cookie_name(self):
        # type: () -> t.Optional[str]
        return self._session_cookie_name

    def get_session_id(self):
        # type: () -> t.Optional[str]
        return self._session_id

    def set_session_id(self, session_id):
        # type: (str) -> None
        self._session_id = session_id

    def remove_session_id(self):
        # type: () -> None
        self._session_id = None

    def _prepare_key(self, key):
        # type: (str) -> str
        if self._session_id:
            if key.startswith(self._prefix):
                key = key[len(self._prefix):]
            return self._prefix + self._session_id + '-' + key
        else:
            if not key.startswith(self._prefix):
                key = self._prefix + key
            return key

    @abstractmethod
    def can_set_keys_expiration(self):
        # type: () -> bool
        raise NotImplementedError

    @abstractmethod
    def get_value(self, key):
        # type: (str) -> T
        raise NotImplementedError

    @abstractmethod
    def set_value(self, key, value, exp=None):
        # type: (str, T, t.Optional[int]) -> None
        raise NotImplementedError

    @abstractmethod
    def check_value(self, key):
        # type: (str) -> bool
        raise NotImplementedError


class DisableSessionId(object):
    _session_id = None  # type: t.Optional[str]
    _launch_data_storage = None  # type: t.Optional[LaunchDataStorage]

    def __init__(self, launch_data_storage):
        # type: (t.Optional[LaunchDataStorage]) -> None
        self._launch_data_storage = launch_data_storage
        if launch_data_storage:
            self._session_id = launch_data_storage.get_session_id()

    def __enter__(self):
        # type: (_T_DISABLED_SESSION_ID) -> _T_DISABLED_SESSION_ID
        if self._launch_data_storage:
            self._launch_data_storage.remove_session_id()
        return self

    def __exit__(self, *args):
        # type: (*t.Any) -> None
        if self._launch_data_storage and self._session_id:
            self._launch_data_storage.set_session_id(self._session_id)



import typing as t

from .base import LaunchDataStorage

T = t.TypeVar('T')


class SessionDataStorage(LaunchDataStorage[T], t.Generic[T]):
    def get_session_cookie_name(self):
        # type: () -> None
        return None

    def set_session_id(self, session_id):
        # type: (str) -> None
        pass

    def get_value(self, key):
        # type: (str) -> T
        assert self._request is not None, 'Request should be set at this point'
        return self._request.session.get(key, None)

    def set_value(self, key, value, exp=None):
        # type: (str, T, t.Optional[int]) -> None
        # pylint: disable=unused-argument
        assert self._request is not None, 'Request should be set at this point'
        self._request.session[key] = value

    def check_value(self, key):
        # type: (str) -> bool
        assert self._request is not None, 'Request should be set at this point'
        return key in self._request.session

    def can_set_keys_expiration(self):
        # type: () -> bool
        return False



# mypy: ignore-errors
from django.contrib import admin

from .models import LtiTool, LtiToolKey


class LtiToolKeyAdmin(admin.ModelAdmin):
    """Admin for LTI Tool Key"""
    list_display = ('id', 'name')

    add_fieldsets = (
        (None, {'fields': ('name', 'private_key', 'public_key')}),
    )

    change_fieldsets = (
        (None, {'fields': ('name', 'private_key', 'public_key', 'public_jwk')}),
    )

    readonly_fields = ('public_jwk',)

    def get_form(self, request, obj=None, **kwargs):  # pylint: disable=arguments-differ
        help_texts = {'public_key_jwk_json': "Tool's generated Public key presented as JWK."}
        kwargs.update({'help_texts': help_texts})
        return super(LtiToolKeyAdmin, self).get_form(request, obj, **kwargs)

    def get_fieldsets(self, request, obj=None):
        if not obj:
            return self.add_fieldsets
        else:
            return self.change_fieldsets


class LtiToolAdmin(admin.ModelAdmin):
    """Admin for LTI Tool"""
    search_fields = ('title', 'issuer', 'client_id', 'auth_login_url', 'auth_token_url', 'key_set_url')
    list_display = ('id', 'title', 'is_active', 'issuer', 'client_id', 'deployment_ids')


admin.site.register(LtiToolKey, LtiToolKeyAdmin)
admin.site.register(LtiTool, LtiToolAdmin)



import typing as t
from abc import ABCMeta, abstractmethod


class CookieService(object):
    __metaclass__ = ABCMeta
    _cookie_prefix = 'lti1p3'  # type: str

    @abstractmethod
    def get_cookie(self, name):
        # type: (str) -> t.Optional[str]
        raise NotImplementedError

    @abstractmethod
    def set_cookie(self, name, value, exp=3600):
        # type: (str, str, int) -> None
        raise NotImplementedError



from abc import ABCMeta
import typing as t

if t.TYPE_CHECKING:
    from typing_extensions import Final


class RoleType(object):
    SYSTEM = 'system'  # type: Final
    INSTITUTION = 'institution'  # type: Final
    CONTEXT = 'membership'  # type: Final


class AbstractRole(object):
    __metaclass__ = ABCMeta
    _base_prefix = 'http://purl.imsglobal.org/vocab/lis/v2'  # type: str
    _role_types = [RoleType.SYSTEM, RoleType.INSTITUTION, RoleType.CONTEXT]  # type: list
    _jwt_roles = []  # type: list
    _common_roles = None  # type: t.Optional[tuple]
    _system_roles = None  # type: t.Optional[tuple]
    _institution_roles = None  # type: t.Optional[tuple]
    _context_roles = None  # type: t.Optional[tuple]

    def __init__(self, jwt_body):
        self._jwt_roles = jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/roles', [])

    def check(self):
        # type: () -> bool
        for role_str in self._jwt_roles:
            role_name, role_type = self.parse_role_str(role_str)
            res = self._check_access(role_name, role_type)
            if res:
                return True
        return False

    def _check_access(self, role_name, role_type=None):
        # type: (str, t.Optional[str]) -> bool
        return bool((self._system_roles and role_type == RoleType.SYSTEM and role_name in self._system_roles)
                    or (self._institution_roles and role_type == RoleType.INSTITUTION
                        and role_name in self._institution_roles)
                    or (self._context_roles and role_type == RoleType.CONTEXT and role_name in self._context_roles)
                    or (self._common_roles and role_type is None and role_name in self._common_roles))

    def parse_role_str(self, role_str):
        # type: (str) -> t.Tuple[str, t.Optional[str]]
        if role_str.startswith(self._base_prefix):
            role = role_str[len(self._base_prefix):]
            role_parts = role.split('/')
            role_name_parts = role.split('#')

            if len(role_parts) > 1 and len(role_name_parts) > 1:
                role_type = role_parts[1]
                role_name = role_name_parts[1]
                if role_type in self._role_types:
                    return role_name, role_type
                return role_name, None
        return role_str, None


class StaffRole(AbstractRole):
    _system_roles = ('Administrator', 'SysAdmin')  # type: tuple
    _institution_roles = ('Faculty', 'SysAdmin', 'Staff', 'Instructor')  # type: tuple


class StudentRole(AbstractRole):
    _common_roles = ('Learner', 'Member', 'User')  # type: tuple
    _system_roles = ('User',)  # type: tuple
    _institution_roles = ('Student', 'Learner', 'Member', 'ProspectiveStudent', 'User')  # type: tuple
    _context_roles = ('Learner', 'Member')  # type: tuple


class TeacherRole(AbstractRole):
    _common_roles = ('Instructor', 'Administrator')  # type: tuple
    _context_roles = ('Instructor', 'Administrator')  # type: tuple


class TeachingAssistantRole(AbstractRole):
    _context_roles = ('TeachingAssistant',)  # type: tuple


class DesignerRole(AbstractRole):
    _common_roles = ('ContentDeveloper',)  # type: tuple
    _context_roles = ('ContentDeveloper',)  # type: tuple


class ObserverRole(AbstractRole):
    _common_roles = ('Mentor',)  # type: tuple
    _context_roles = ('Mentor',)  # type: tuple



from abc import ABCMeta, abstractmethod
from ..exception import LtiException


class MessageValidatorAbstract(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def validate(self, jwt_body):
        raise NotImplementedError

    @abstractmethod
    def can_validate(self, jwt_body):
        raise NotImplementedError

    def run_common_validators(self, jwt_body):
        if not jwt_body.get('sub'):
            raise LtiException('Must have a user (sub)')

        if jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/version') != '1.3.0':
            raise LtiException('Incorrect version, expected 1.3.0')

        if not jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/roles'):
            raise LtiException('Missing Roles Claim')



import typing as t

from ..deployment import Deployment
from ..registration import Registration
from ..request import Request
from .abstract import ToolConfAbstract

if t.TYPE_CHECKING:
    from ..message_launch import _LaunchData
    from typing_extensions import Literal
    from .abstract import FIND_REG_KWARGS


class ToolConfDict(ToolConfAbstract[Request]):
    _config = None
    _private_key = None  # type: t.Optional[t.Mapping[str, str]]
    _public_key = None  # type: t.Optional[t.Mapping[str, str]]

    def __init__(self, json_data):
        # type: (dict) -> None
        """
        json_data is a dict where each key is issuer and value is issuer's configuration.
        Configuration could be set in two formats:

        1. { ... "iss": { ... "client_id: "client" ... }, ... }
        In this case the library will work in the concept: one issuer ~ one client-id

        2. { ... "iss": [ { ... "client_id: "client1" ... }, { ... "client_id: "client2" ... } ], ... }
        In this case the library will work in concept: one issuer ~ many client-ids

        Example:
            {
                "iss1": [{
                        "default": True,
                        "client_id": "client_id1",
                        "auth_login_url": "auth_login_url1",
                        "auth_token_url": "auth_token_url1",
                        "auth_audience": None,
                        "key_set_url": "key_set_url1",
                        "key_set": None,
                        "deployment_ids": ["deployment_id1", "deployment_id2"]
                    }, {
                        "default": False,
                        "client_id": "client_id2",
                        "auth_login_url": "auth_login_url2",
                        "auth_token_url": "auth_token_url2",
                        "auth_audience": None,
                        "key_set_url": "key_set_url2",
                        "key_set": None,
                        "deployment_ids": ["deployment_id3", "deployment_id4"]
                    }],
                "iss2": [ .... ]
            }

        default (bool) - this iss config will be used in case if client-id was not passed on the login step
        client_id - this is the id received in the 'aud' during a launch
        auth_login_url - the platform's OIDC login endpoint
        auth_token_url - the platform's service authorization endpoint
        auth_audience - the platform's OAuth2 Audience (aud). Is used to get platform's access token,
                        Usually the same as "auth_token_url" but in the common case could be a different url
        key_set_url - the platform's JWKS endpoint
        key_set - in case if platform's JWKS endpoint somehow unavailable you may paste JWKS here
        deployment_ids (list) - The deployment_id passed by the platform during launch
        """
        super(ToolConfDict, self).__init__()
        if not isinstance(json_data, dict):
            raise Exception("Invalid tool conf format. Must be dict")

        for iss, iss_conf in json_data.items():
            if isinstance(iss_conf, dict):
                self.set_iss_has_one_client(iss)
                self._validate_iss_config_item(iss, iss_conf)
            elif isinstance(iss_conf, list):
                self.set_iss_has_many_clients(iss)
                for v in iss_conf:
                    self._validate_iss_config_item(iss, v)
            else:
                raise Exception("Invalid tool conf format. Allowed types of elements: list or dict")

        self._config = json_data
        self._private_key = {}
        self._public_key = {}

    def _validate_iss_config_item(self, iss, iss_config_item):
        # type: (str, t.Any) -> None
        if not isinstance(iss_config_item, dict):
            raise Exception("Invalid configuration %s for the %s issuer. Must be dict" % (iss, str(iss_config_item)))
        required_keys = ['auth_login_url', 'auth_token_url', 'client_id', 'deployment_ids']
        for key in required_keys:
            if key not in iss_config_item:
                raise Exception("Key '%s' is missing in the %s config for the %s issuer"
                                % (key, str(iss_config_item), iss))
        if not isinstance(iss_config_item['deployment_ids'], list):
            raise Exception("Invalid deployment_ids value in the %s config for the %s issuer. Must be a list"
                            % (str(iss_config_item), iss))

    def _get_registration(self, iss, iss_conf):
        # type: (str, t.Any) -> Registration
        reg = Registration()
        reg.set_auth_login_url(iss_conf['auth_login_url'])\
            .set_auth_token_url(iss_conf['auth_token_url'])\
            .set_client_id(iss_conf['client_id'])\
            .set_key_set(iss_conf.get('key_set'))\
            .set_key_set_url(iss_conf.get('key_set_url'))\
            .set_issuer(iss)\
            .set_tool_private_key(self.get_private_key(iss, iss_conf['client_id']))
        auth_audience = iss_conf.get('auth_audience')
        if auth_audience:
            reg.set_auth_audience(auth_audience)
        public_key = self.get_public_key(iss, iss_conf['client_id'])
        if public_key:
            reg.set_tool_public_key(public_key)
        return reg

    def _get_deployment(self, iss_conf, deployment_id):
        if deployment_id not in iss_conf['deployment_ids']:
            return None
        d = Deployment()
        return d.set_deployment_id(deployment_id)

    def find_registration_by_issuer(self, iss, *args, **kwargs):
        # pylint: disable=unused-argument
        iss_conf = self.get_iss_config(iss)
        return self._get_registration(iss, iss_conf)

    def find_registration_by_params(self, iss, client_id, *args, **kwargs):
        # pylint: disable=unused-argument
        iss_conf = self.get_iss_config(iss, client_id)
        return self._get_registration(iss, iss_conf)

    def find_deployment(self, iss, deployment_id):
        iss_conf = self.get_iss_config(iss)
        return self._get_deployment(iss_conf, deployment_id)

    def find_deployment_by_params(self, iss, deployment_id, client_id, *args, **kwargs):
        # pylint: disable=unused-argument
        iss_conf = self.get_iss_config(iss, client_id)
        return self._get_deployment(iss_conf, deployment_id)

    def set_public_key(self, iss, key_content, client_id=None):
        if self.check_iss_has_many_clients(iss):
            if not client_id:
                raise Exception("Can't set public key: missing client_id")
            if iss not in self._public_key:
                self._public_key[iss] = {}
            self._public_key[iss][client_id] = key_content
        else:
            self._public_key[iss] = key_content

    def get_public_key(self, iss, client_id=None):
        if self.check_iss_has_many_clients(iss):
            if not client_id:
                raise Exception("Can't get public key: missing client_id")
            return self._public_key.get(iss, {}).get(client_id)
        else:
            return self._public_key.get(iss)

    def set_private_key(self, iss, key_content, client_id=None):
        if self.check_iss_has_many_clients(iss):
            if not client_id:
                raise Exception("Can't set private key: missing client_id")
            if iss not in self._private_key:
                self._private_key[iss] = {}
            self._private_key[iss][client_id] = key_content
        else:
            self._private_key[iss] = key_content

    def get_private_key(self, iss, client_id=None):
        if self.check_iss_has_many_clients(iss):
            if not client_id:
                raise Exception("Can't get private key: missing client_id")
            return self._private_key.get(iss, {}).get(client_id)
        else:
            return self._private_key.get(iss)

    def get_iss_config(self, iss, client_id=None):
        if iss not in self._config:
            raise Exception('iss %s not found in settings' % iss)

        if isinstance(self._config[iss], list):
            items_len = len(self._config[iss])
            for subitem in self._config[iss]:
                if (client_id and subitem['client_id'] == client_id)\
                        or (not client_id and subitem.get('default', False))\
                        or (not client_id and items_len == 1):
                    return subitem
            raise Exception('iss %s [client_id=%s] not found in settings' % (iss, client_id))
        return self._config[iss]

    def get_jwks(self, iss=None, client_id=None, **kwargs):
        # pylint: disable=unused-argument
        if iss or client_id:
            return super(ToolConfDict, self).get_jwks(iss, client_id)

        public_keys = []
        for iss_item in self._public_key.values():
            if isinstance(iss_item, dict):
                for pub_key in iss_item.values():
                    if pub_key not in public_keys:
                        public_keys.append(pub_key)
            else:
                if iss_item not in public_keys:
                    public_keys.append(iss_item)
        return {
            'keys': [Registration.get_jwk(k) for k in public_keys]
        }



import sys

if sys.version_info > (2, ):
    def encode_on_py3(arg, encoding):
        # type: (str, str) -> bytes
        return arg.encode(encoding)
else:
    def encode_on_py3(arg, encoding):
        # type: (str, str) -> bytes
        # pylint: disable=unused-argument
        return arg



import hashlib
import sys
import time
import typing as t
import uuid

import jwt  # type: ignore
import requests

from .exception import LtiException

if t.TYPE_CHECKING:
    from mypy_extensions import TypedDict
    from .registration import Registration

    _ServiceConnectorResponse = TypedDict('_ServiceConnectorResponse', {
        'headers': t.Dict[str, str],
        'body': t.Union[None, int, float, t.List[object], t.Dict[str, object], str],
    })


class ServiceConnector(object):
    _registration = None  # type: Registration
    _access_tokens = None  # type: t.Dict[str, str]

    def __init__(self, registration):
        # type: (Registration) -> None
        self._registration = registration
        self._access_tokens = {}

    def get_access_token(self, scopes):
        # type: (t.Sequence[str]) -> str

        # Don't fetch the same key more than once
        scopes = sorted(scopes)
        scopes_str = '|'.join(scopes)  # type: str

        if sys.version_info[0] > 2:
            scopes_bytes = scopes_str.encode('utf-8')
        else:
            scopes_bytes = scopes_str
        scope_key = hashlib.md5(scopes_bytes).hexdigest()

        if scope_key in self._access_tokens:
            return self._access_tokens[scope_key]

        # Build up JWT to exchange for an auth token
        client_id = self._registration.get_client_id()
        auth_url = self._registration.get_auth_token_url()
        assert auth_url is not None, 'auth_url should be set at this point'
        auth_audience = self._registration.get_auth_audience()
        aud = auth_audience if auth_audience else auth_url

        jwt_claim = {
            "iss": client_id,
            "sub": client_id,
            "aud": aud,
            "iat": int(time.time()) - 5,
            "exp": int(time.time()) + 60,
            "jti": 'lti-service-token-' + str(uuid.uuid4())
        }
        headers = None
        kid = self._registration.get_kid()
        if kid:
            headers = {'kid': kid}

        # Sign the JWT with our private key (given by the platform on registration)
        private_key = self._registration.get_tool_private_key()
        assert private_key is not None, 'Private key should be set at this point'
        jwt_val = jwt.encode(jwt_claim, private_key, algorithm='RS256',
                             headers=headers)

        auth_request = {
            'grant_type': 'client_credentials',
            'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': jwt_val,
            'scope': ' '.join(scopes)
        }

        # Make request to get auth token
        r = requests.post(auth_url, data=auth_request)
        if not r.ok:
            raise LtiException('HTTP response [%s]: %s - %s' % (auth_url, str(r.status_code), r.text))
        response = r.json()

        self._access_tokens[scope_key] = response['access_token']
        return self._access_tokens[scope_key]

    def make_service_request(
            self,
            scopes,  # type: t.Sequence[str]
            url,  # type: str
            is_post=False,  # type: bool
            data=None,  # type: t.Union[None, str]
            content_type='application/json',  # type: str
            accept='application/json',  # type: str
    ):
        # type: (...) -> _ServiceConnectorResponse
        access_token = self.get_access_token(scopes)
        headers = {
            'Authorization': 'Bearer ' + access_token,
            'Accept': accept
        }

        if is_post:
            headers['Content-Type'] = content_type
            post_data = data or None
            r = requests.post(url, data=post_data, headers=headers)
        else:
            r = requests.get(url, headers=headers)

        if r.status_code not in (200, 201):
            raise LtiException('HTTP response [%s]: %s - %s' % (url, str(r.status_code), r.text))

        return {
            'headers': dict(r.headers),
            'body': r.json() if r.content else None
        }



# mypy: ignore-errors
import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LtiToolKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Key name', max_length=255, unique=True)),
                ('private_key', models.TextField(help_text="Tool's generated Private key. "
                                                           "Keep this value in secret")),
                ('public_key', models.TextField(blank=True, help_text="Tool's generated Public key", null=True)),
                ('public_jwk', models.TextField(blank=True, help_text="Tool's generated Public key (from the "
                                                                      "field above) presented as JWK.", null=True)),
            ],
            options={
                'verbose_name': 'lti 1.3 tool key',
                'verbose_name_plural': 'lti 1.3 tool keys',
                'db_table': 'lti1p3_tool_key',
            },
        ),
        migrations.CreateModel(
            name='LtiTool',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('is_active', models.BooleanField(default=True)),
                ('issuer', models.CharField(help_text="This will usually look something like 'http://example.com'. "
                                                      "Value provided by LTI 1.3 Platform", max_length=255)),
                ('client_id', models.CharField(help_text='Value provided by LTI 1.3 Platform', max_length=255)),
                ('use_by_default', models.BooleanField(default=False, help_text='This iss config will be used in case '
                                                                                'if client-id was not passed')),
                ('auth_login_url', models.CharField(help_text="The platform's OIDC login endpoint. "
                                                              "Value provided by LTI 1.3 Platform", max_length=1024,
                                                    validators=[django.core.validators.URLValidator()])),
                ('auth_token_url', models.CharField(help_text="The platform's service authorization endpoint. "
                                                              "Value provided by LTI 1.3 Platform", max_length=1024,
                                                    validators=[django.core.validators.URLValidator()])),
                ('auth_audience', models.CharField(blank=True, help_text="The platform's OAuth2 Audience (aud). "
                                                                         "Usually could be skipped", max_length=1024,
                                                   null=True)),
                ('key_set_url', models.CharField(blank=True, help_text="The platform's JWKS endpoint. "
                                                                       "Value provided by LTI 1.3 Platform",
                                                 max_length=1024, null=True,
                                                 validators=[django.core.validators.URLValidator()])),
                ('key_set', models.TextField(blank=True, help_text="In case if platform's JWKS endpoint somehow "
                                                                   "unavailable you may paste JWKS here. Value "
                                                                   "provided by LTI 1.3 Platform", null=True)),
                ('deployment_ids', models.TextField(help_text='List of Deployment IDs. Example: '
                                                              '["test-id-1", "test-id-2", ...] Each value '
                                                              'is provided by LTI 1.3 Platform. ')),
                ('tool_key', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='lti_tools',
                                               to='lti1p3_tool_config.LtiToolKey')),
            ],
            options={
                'verbose_name': 'lti 1.3 tool',
                'verbose_name_plural': 'lti 1.3 tools',
                'db_table': 'lti1p3_tool',
                'unique_together': {('issuer', 'client_id')},
            },
        ),
    ]



import re
import typing as t

if t.TYPE_CHECKING:
    from mypy_extensions import TypedDict
    from .service_connector import ServiceConnector
    from typing_extensions import Literal

    _NamesAndRolesData = TypedDict('_NamesAndRolesData', {
        'context_memberships_url': str,
    }, total=False)
    _Member = TypedDict('_Member', {
        'name': str,
        'status': Literal['Active', 'Inactive', 'Deleted'],
        'picture': str,
        'given_name': str,
        'family_name': str,
        'middle_name': str,
        'email': str,
        'user_id': str,
        'lis_person_sourcedid': str,
        'roles': t.List[str],
        'message': t.Union[t.List[t.Dict[str, object]], t.Dict[str, object]],
        'lti11_legacy_user_id': t.Optional[str],
    }, total=False)


class NamesRolesProvisioningService(object):
    _service_connector = None  # type: ServiceConnector
    _service_data = None  # type: _NamesAndRolesData

    def __init__(self, service_connector, service_data):
        # type: (ServiceConnector, _NamesAndRolesData) -> None
        self._service_connector = service_connector
        self._service_data = service_data

    def get_members(self):
        # type: () -> t.List[_Member]
        members = []  # type: t.List[_Member]
        next_page = self._service_data['context_memberships_url']  # type: t.Union[Literal[False], str]

        while next_page:
            page = self._service_connector.make_service_request(
                ['https://purl.imsglobal.org/spec/lti-nrps/scope/contextmembership.readonly'],
                next_page,  # type: ignore
                accept='application/vnd.ims.lti-nrps.v2.membershipcontainer+json'
            )

            members.extend(t.cast(t.Any, page.get('body', {})).get('members', []))

            next_page = False
            link_header = page.get('headers', {}).get('link', '')
            if link_header:
                match = re.search(r'<([^>]*)>;\s*rel="next"', link_header.replace('\n', ' ').lower().strip())
                if match:
                    next_page = match.group(1)

        return members



import json

from pylti1p3.deployment import Deployment
from pylti1p3.exception import LtiException
from pylti1p3.registration import Registration
from pylti1p3.tool_config.abstract import ToolConfAbstract


default_app_config = 'pylti1p3.contrib.django.lti1p3_tool_config.apps.PyLTI1p3ToolConfig'


class DjangoDbToolConf(ToolConfAbstract):
    _lti_tools = None
    _tools_cls = None
    _keys_cls = None

    def __init__(self):
        from .models import LtiTool, LtiToolKey
        super(DjangoDbToolConf, self).__init__()
        self._lti_tools = {}
        self._tools_cls = LtiTool
        self._keys_cls = LtiToolKey

    def get_lti_tool(self, iss, client_id):
        lti_tool = self._lti_tools.get(iss) if client_id is None else self._lti_tools.get(iss, {}).get(client_id)
        if lti_tool:
            return lti_tool

        if client_id is None:
            lti_tool = self._tools_cls.objects.filter(issuer=iss, is_active=True).order_by('use_by_default').first()
        else:
            try:
                lti_tool = self._tools_cls.objects.get(issuer=iss, client_id=client_id, is_active=True)
            except self._tools_cls.DoesNotExist:
                pass

        if lti_tool is None:
            raise LtiException('iss %s [client_id=%s] not found in settings' % (iss, client_id))

        if client_id is None:
            self._lti_tools[iss] = lti_tool
        else:
            if iss not in self._lti_tools:
                self._lti_tools[iss] = {}
            self._lti_tools[iss][client_id] = lti_tool

        return lti_tool

    def check_iss_has_one_client(self, iss):
        return False

    def check_iss_has_many_clients(self, iss):
        return True

    def find_registration_by_issuer(self, iss, *args, **kwargs):
        pass

    def find_registration_by_params(self, iss, client_id, *args, **kwargs):
        lti_tool = self.get_lti_tool(iss, client_id)
        auth_audience = lti_tool.auth_audience if lti_tool.auth_audience else None
        key_set = json.loads(lti_tool.key_set) if lti_tool.key_set else None
        key_set_url = lti_tool.key_set_url if lti_tool.key_set_url else None
        tool_public_key = lti_tool.tool_key.public_key if lti_tool.tool_key.public_key else None

        reg = Registration()
        reg.set_auth_login_url(lti_tool.auth_login_url) \
            .set_auth_token_url(lti_tool.auth_token_url) \
            .set_auth_audience(auth_audience) \
            .set_client_id(lti_tool.client_id) \
            .set_key_set(key_set) \
            .set_key_set_url(key_set_url) \
            .set_issuer(lti_tool.issuer) \
            .set_tool_private_key(lti_tool.tool_key.private_key) \
            .set_tool_public_key(tool_public_key)
        return reg

    def find_deployment(self, iss, deployment_id):
        pass

    def find_deployment_by_params(self, iss, deployment_id, client_id, *args, **kwargs):
        lti_tool = self.get_lti_tool(iss, client_id)
        deployment_ids = json.loads(lti_tool.deployment_ids) if lti_tool.deployment_ids else []
        if deployment_id not in deployment_ids:
            return None
        d = Deployment()
        return d.set_deployment_id(deployment_id)

    def get_jwks(self, iss=None, client_id=None, **kwargs):
        search_kwargs = {}
        if iss:
            search_kwargs['lti_tools__issuer'] = iss
        if client_id:
            search_kwargs['lti_tools__client_id'] = client_id

        if search_kwargs:
            search_kwargs['lti_tools__is_active'] = True
            qs = self._keys_cls.objects.filter(**search_kwargs)
        else:
            qs = self._keys_cls.objects.all()

        jwks = []
        public_key_lst = []

        for key in qs:
            if key.public_key and key.public_key not in public_key_lst:
                if key.public_jwk:
                    jwks.append(json.loads(key.public_jwk))
                else:
                    jwks.append(Registration.get_jwk(key.public_key))
                public_key_lst.append(key.public_key)
        return {
            'keys': jwks
        }



import json
import typing as t

if t.TYPE_CHECKING:
    T_SELF = t.TypeVar('T_SELF', bound='Grade')
    EXTRA_CLAIMS = t.Mapping[str, t.Any]


class Grade(object):
    _score_given = None  # type: t.Optional[float]
    _score_maximum = None  # type: t.Optional[float]
    _activity_progress = None  # type: t.Optional[str]
    _grading_progress = None  # type: t.Optional[str]
    _timestamp = None  # type: t.Optional[str]
    _user_id = None  # type: t.Optional[str]
    _extra_claims = None  # type: t.Optional[EXTRA_CLAIMS]

    def get_score_given(self):
        # type: () -> t.Optional[float]
        return self._score_given

    def set_score_given(self, value):
        # type: (T_SELF, float) -> T_SELF
        self._score_given = value
        return self

    def get_score_maximum(self):
        # type: () -> t.Optional[float]
        return self._score_maximum

    def set_score_maximum(self, value):
        # type: (T_SELF, float) -> T_SELF
        self._score_maximum = value
        return self

    def get_activity_progress(self):
        # type: () -> t.Optional[str]
        return self._activity_progress

    def set_activity_progress(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._activity_progress = value
        return self

    def get_grading_progress(self):
        # type: () -> t.Optional[str]
        return self._grading_progress

    def set_grading_progress(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._grading_progress = value
        return self

    def get_timestamp(self):
        # type: () -> t.Optional[str]
        return self._timestamp

    def set_timestamp(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._timestamp = value
        return self

    def get_user_id(self):
        # type: () -> t.Optional[str]
        return self._user_id

    def set_user_id(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._user_id = value
        return self

    def set_extra_claims(self, value):
        # type: (T_SELF, EXTRA_CLAIMS) -> T_SELF
        self._extra_claims = value
        return self

    def get_extra_claims(self):
        # type: () -> t.Optional[EXTRA_CLAIMS]
        return self._extra_claims

    def get_value(self):
        # type: () -> str
        data = {
            'scoreGiven': self._score_given if self._score_given else None,
            'scoreMaximum': self._score_maximum if self._score_maximum else None,
            'activityProgress': self._activity_progress,
            'gradingProgress': self._grading_progress,
            'timestamp': self._timestamp,
            'userId': self._user_id
        }
        if self._extra_claims is not None:
            data.update(self._extra_claims)

        return json.dumps({k: v for k, v in data.items() if v is not None})



import warnings
import werkzeug

from pylti1p3.cookie import CookieService


class FlaskCookieService(CookieService):
    _request = None
    _cookie_data_to_set = None

    def __init__(self, request):
        self._request = request
        self._cookie_data_to_set = {}

    def _get_key(self, key):
        return self._cookie_prefix + '-' + key

    def get_cookie(self, name):
        return self._request.get_cookie(self._get_key(name))

    def set_cookie(self, name, value, exp=3600):
        self._cookie_data_to_set[self._get_key(name)] = {
            'value': value,
            'exp': exp
        }

    def update_response(self, response):
        warning_raised = False
        werkzeug_version = int(werkzeug.__version__.split('.')[0])

        for key, cookie_data in self._cookie_data_to_set.items():
            cookie_kwargs = dict(
                key=key,
                value=cookie_data['value'],
                max_age=cookie_data['exp'],
                secure=self._request.is_secure(),
                path='/',
                httponly=True,
            )

            if werkzeug_version >= 1 and self._request.is_secure():
                cookie_kwargs['samesite'] = 'None'
            elif werkzeug_version < 1 and not warning_raised:
                warnings.warn("The samesite argument is not allowed for werkzeug less than 1.0. "
                              "So there is no guarantee that PyLTI1p3 cookies will be set inside the iframes. "
                              "Please update werkzeug", Warning)
                warning_raised = True

            response.set_cookie(**cookie_kwargs)



from .cookie import FlaskCookieService
from .oidc_login import FlaskOIDCLogin
from .message_launch import FlaskMessageLaunch
from .request import FlaskRequest
from .session import FlaskSessionService
from .launch_data_storage.cache import FlaskCacheDataStorage



import typing as t

if t.TYPE_CHECKING:
    from typing_extensions import Final


class Action(object):
    OIDC_LOGIN = 'oidc_login'  # type: Final
    MESSAGE_LAUNCH = 'message_launch'  # type: Final



import typing as t

from .exception import LtiException
from .lineitem import LineItem

if t.TYPE_CHECKING:
    from .service_connector import ServiceConnector, _ServiceConnectorResponse
    from .grade import Grade
    from mypy_extensions import TypedDict
    from typing_extensions import Literal

    _AssignmentsGradersData = TypedDict('_AssignmentsGradersData', {
        'scope': t.List[Literal['https://purl.imsglobal.org/spec/lti-ags/scope/score',
                                'https://purl.imsglobal.org/spec/lti-ags/scope/lineitem',
                                'https://purl.imsglobal.org/spec/lti-ags/scope/lineitem.readonly']],
        'lineitems': str,
        'lineitem': str,
    }, total=False)


class AssignmentsGradesService(object):
    _service_connector = None  # type: ServiceConnector
    _service_data = None  # type: _AssignmentsGradersData

    def __init__(self, service_connector, service_data):
        # type: (ServiceConnector, _AssignmentsGradersData) -> None
        self._service_connector = service_connector
        self._service_data = service_data

    def put_grade(self, grade, line_item=None):
        # type: (Grade, t.Optional[LineItem]) -> _ServiceConnectorResponse
        if "https://purl.imsglobal.org/spec/lti-ags/scope/score" not in self._service_data['scope']:
            raise LtiException('Missing required scope')

        if line_item and not line_item.get_id():
            line_item = self.find_or_create_lineitem(line_item)
            score_url = line_item.get_id()
        elif not line_item and self._service_data.get('lineitem'):
            score_url = self._service_data.get('lineitem')
        else:
            if not line_item:
                line_item = LineItem()
                line_item.set_label('default')\
                    .set_score_maximum(100)
                line_item = self.find_or_create_lineitem(line_item)
            score_url = line_item.get_id()

        assert score_url is not None
        score_url = self._add_url_path_ending(score_url, 'scores')
        return self._service_connector.make_service_request(
            self._service_data['scope'],
            score_url,
            is_post=True,
            data=grade.get_value(),
            content_type='application/vnd.ims.lis.v1.score+json'
        )

    def get_lineitems(self):
        # type: () -> list
        if "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem" not in self._service_data['scope']:
            raise LtiException('Missing required scope')

        line_items = self._service_connector.make_service_request(
            self._service_data['scope'],
            self._service_data['lineitems'],
            accept='application/vnd.ims.lis.v2.lineitemcontainer+json'
        )
        if not isinstance(line_items['body'], list):
            raise LtiException('Unknown response type received for line items')
        return line_items['body']

    def find_lineitem_by_id(self, ln_id):
        # type: (t.Optional[str]) -> t.Optional[LineItem]
        line_items = self.get_lineitems()

        for line_item in line_items:
            line_item_id = line_item.get('id')
            if line_item_id == ln_id:
                return LineItem(line_item)
        return None

    def find_lineitem_by_tag(self, tag):
        # type: (t.Optional[str]) -> t.Optional[LineItem]
        line_items = self.get_lineitems()

        for line_item in line_items:
            line_item_tag = line_item.get('tag')
            if line_item_tag == tag:
                return LineItem(line_item)
        return None

    def find_or_create_lineitem(self, new_line_item, find_by='tag'):
        # type: (LineItem, Literal['tag', 'id']) -> LineItem
        if find_by == 'tag':
            tag = new_line_item.get_tag()
            line_item = self.find_lineitem_by_tag(tag)
        elif find_by == 'id':
            line_id = new_line_item.get_id()
            line_item = self.find_lineitem_by_id(line_id)
        else:
            raise LtiException('Invalid "find_by" value: ' + str(find_by))

        if line_item:
            return line_item

        created_line_item = self._service_connector.make_service_request(
            self._service_data['scope'],
            self._service_data['lineitems'],
            is_post=True,
            data=new_line_item.get_value(),
            content_type='application/vnd.ims.lis.v2.lineitem+json',
            accept='application/vnd.ims.lis.v2.lineitem+json'
        )
        if not isinstance(created_line_item['body'], dict):
            raise LtiException('Unknown response type received for create line item')
        return LineItem(created_line_item['body'])

    def get_grades(self, line_item):
        # type: (LineItem) -> list
        line_item_id = line_item.get_id()
        line_item_tag = line_item.get_tag()

        find_by = None  # type: t.Optional[Literal['id', 'tag']]
        if line_item_id:
            find_by = 'id'
        elif line_item_tag:
            find_by = 'tag'
        else:
            raise LtiException('Received LineItem did not contain a tag or id')

        line_item = self.find_or_create_lineitem(line_item, find_by=find_by)
        line_item_id = line_item.get_id()
        assert line_item_id is not None
        results_url = self._add_url_path_ending(line_item_id, 'results')
        scores = self._service_connector.make_service_request(
            self._service_data['scope'],
            results_url,
            accept='application/vnd.ims.lis.v2.resultcontainer+json'
        )
        if not isinstance(scores['body'], list):
            raise LtiException('Unknown response type received for results')
        return scores['body']

    def _add_url_path_ending(self, url, url_path_ending):
        # type: (str, str) -> str
        if '?' in url:
            url_parts = url.split('?')
            new_url = url_parts[0]
            new_url += '' if new_url.endswith('/') else '/'
            return new_url + url_path_ending + '?' + url_parts[1]
        else:
            url += '' if url.endswith('/') else '/'
            return url + url_path_ending



from django.core.cache import caches  # type: ignore
from pylti1p3.launch_data_storage.cache import CacheDataStorage


class DjangoCacheDataStorage(CacheDataStorage):
    _cache = None

    def __init__(self, cache_name='default', **kwargs):
        self._cache = caches[cache_name]
        super(DjangoCacheDataStorage, self).__init__(cache_name, **kwargs)



import sys

import django  # type: ignore
from pylti1p3.cookie import CookieService

if sys.version_info[0] > 2:
    import http.cookies as Cookie  # type: ignore
else:
    import Cookie  # type: ignore

# Add support for the SameSite attribute (obsolete when PY37 is unsupported).
# pylint: disable=protected-access
if 'samesite' not in Cookie.Morsel._reserved:  # type: ignore
    Cookie.Morsel._reserved.setdefault('samesite', 'SameSite')  # type: ignore


class DjangoCookieService(CookieService):
    _request = None
    _cookie_data_to_set = None

    def __init__(self, request):
        self._request = request
        self._cookie_data_to_set = {}

    def _get_key(self, key):
        return self._cookie_prefix + '-' + key

    def get_cookie(self, name):
        return self._request.get_cookie(self._get_key(name))

    def set_cookie(self, name, value, exp=3600):
        self._cookie_data_to_set[self._get_key(name)] = {
            'value': value,
            'exp': exp,
        }

    def update_response(self, response):
        for key, cookie_data in self._cookie_data_to_set.items():
            kwargs = {
                'value': cookie_data['value'],
                'max_age': cookie_data['exp'],
                'secure': self._request.is_secure(),
                'httponly': True,
                'path': '/'
            }

            if self._request.is_secure():
                # samesite argument was added in Django 2.1, but samesite could be set as None only from Django 3.1
                # https://github.com/django/django/pull/11894
                django_support_samesite_none = django.VERSION[0] > 3 \
                                               or (django.VERSION[0] == 3 and django.VERSION[1] >= 1)

                # SameSite=None and Secure=True are required to work inside iframes
                if django_support_samesite_none:
                    kwargs['samesite'] = 'None'
                    response.set_cookie(key, **kwargs)
                else:
                    response.set_cookie(key, **kwargs)
                    response.cookies[key]['samesite'] = 'None'
            else:
                response.set_cookie(key, **kwargs)



import typing as t

from .base import LaunchDataStorage

T = t.TypeVar('T')


class CacheDataStorage(LaunchDataStorage[T], t.Generic[T]):
    # This seems a bit strange, I have no idea when this is ever set.
    _cache = None  # type: t.Any

    def get_session_cookie_name(self):
        # type: () -> t.Optional[str]
        """
        Workaround for the local non-HTTP usage.
        There is odd situation that all cookies become unavailable from some time
        on the launch step (even if they were set in the new window).
        Looks like it is bug in Chrome >= 80 related to SameSite changes.
        So because of this we have to set all cache values without session prefix.
        It is less secure because if you know unique launch_id you may get access to launch data,
        but unfortunately there is no other way. So please use HTTPS on production :-)
        """
        assert self._request is not None, 'Request should be set at this point'
        if not self._request.is_secure():
            return None
        return super(CacheDataStorage, self).get_session_cookie_name()

    def get_value(self, key):
        # type: (str) -> T
        key = self._prepare_key(key)
        return self._cache.get(key)

    def set_value(self, key, value, exp=None):
        # type: (str, T, t.Optional[int]) -> None
        key = self._prepare_key(key)
        self._cache.set(key, value, exp)

    def check_value(self, key):
        # type: (str) -> bool
        key = self._prepare_key(key)
        return self._cache.get(key) is not None

    def can_set_keys_expiration(self):
        # type: () -> bool
        return True



import typing as t

if t.TYPE_CHECKING:
    T_SELF = t.TypeVar('T_SELF', bound='Deployment')


class Deployment(object):

    _deployment_id = None  # type: t.Optional[str]

    def get_deployment_id(self):
        # type: () -> t.Optional[str]
        return self._deployment_id

    def set_deployment_id(self, deployment_id):
        # type: (T_SELF, str) -> T_SELF
        self._deployment_id = deployment_id
        return self



import base64
import hashlib
import json
import string  # pylint: disable=deprecated-module
import typing as t
import uuid
from abc import ABCMeta, abstractmethod

import jwt  # type: ignore
import requests
from jwcrypto.jwk import JWK  # type: ignore

from .actions import Action
from .assignments_grades import AssignmentsGradesService
from .deep_link import DeepLink
from .exception import LtiException
from .launch_data_storage.base import DisableSessionId
from .message_validators import get_validators
from .names_roles import NamesRolesProvisioningService
from .roles import StaffRole, StudentRole, TeacherRole, TeachingAssistantRole, DesignerRole, ObserverRole
from .service_connector import ServiceConnector
from .utils import encode_on_py3

if t.TYPE_CHECKING:
    from .launch_data_storage.base import LaunchDataStorage
    from .registration import Registration, _KeySet
    from .request import Request
    from .tool_config import ToolConfAbstract  # pylint: disable=cyclic-import
    from .session import SessionService
    from .cookie import CookieService
    from .deep_link import _DeepLinkData
    from .names_roles import _NamesAndRolesData
    from .assignments_grades import _AssignmentsGradersData

    from mypy_extensions import TypedDict
    from typing_extensions import Literal

    _ResourceLinkClaim = TypedDict('_ResourceLinkClaim', {
        # Required data
        'id': str,

        # Optional data
        'description': str,
        'title': str,
    }, total=False)

    _ContextClaim = TypedDict('_ContextClaim', {
        # Required data
        'id': str,

        # Optional data
        'label': str,
        'title': str,
        'type': t.List[str],
    }, total=False)

    _ToolPlatformClaim = TypedDict('_ToolPlatformClaim', {
        # Required data
        'guid': str,

        # Optional data
        'contact_email': str,
        'description': str,
        'name': str,
        'url': str,
        'product_family_code': str,
        'version': str,
    }, total=False)

    _LearningInformationServicesClaim = TypedDict('_LearningInformationServicesClaim', {
        'person_sourcedid': str,
        'course_offering_sourcedid': str,
        'course_section_sourcedid': str,
    }, total=False)

    _MigrationClaim = TypedDict(
        '_MigrationClaim', {
            # Required data
            'oauth_consumer_key': str,

            # Optional data
            'oauth_consumer_key_sign': str,
            'user_id': str,
            'context_id': str,
            'tool_consumer_instance_guid ': str,
            'resource_link_id': str,
        },
        total=False,
    )

    _LaunchData = TypedDict(
        '_LaunchData', {
            # Required data
            'iss': str,
            'nonce': str,
            'aud': t.Union[t.List[str], str],
            'https://purl.imsglobal.org/spec/lti/claim/message_type':
                Literal['LtiResourceLinkRequest', 'LtiDeepLinkingRequest'],
            'https://purl.imsglobal.org/spec/lti/claim/version': Literal['1.3.0'],
            'https://purl.imsglobal.org/spec/lti/claim/deployment_id': str,
            'https://purl.imsglobal.org/spec/lti/claim/target_link_uri': str,
            'https://purl.imsglobal.org/spec/lti/claim/resource_link': _ResourceLinkClaim,
            'https://purl.imsglobal.org/spec/lti/claim/roles': t.List[str],
            'sub': str,


            # Optional data
            'given_name': str,
            'family_name': str,
            'name': str,
            'email': str,
            'https://purl.imsglobal.org/spec/lti/claim/context': _ContextClaim,
            'https://purl.imsglobal.org/spec/lti/claim/lis': _LearningInformationServicesClaim,
            'https://purl.imsglobal.org/spec/lti/claim/custom': t.Mapping[str, str],
            'https://purl.imsglobal.org/spec/lti-dl/claim/deep_linking_settings': _DeepLinkData,
            'https://purl.imsglobal.org/spec/lti-nrps/claim/namesroleservice': _NamesAndRolesData,
            'https://purl.imsglobal.org/spec/lti-ags/claim/endpoint': _AssignmentsGradersData,
            'https://purl.imsglobal.org/spec/lti/claim/tool_platform': _ToolPlatformClaim,
            'https://purl.imsglobal.org/spec/lti/claim/role_scope_mentor': t.List[str],
            'https://purl.imsglobal.org/spec/lti/claim/lti1p1': _MigrationClaim,
        },
        total=False
    )

    _JwtHeader = TypedDict(
        '_JwtHeader', {
            'kid': str,
            'alg': str,
        },
        total=False,
    )

    _JwtData = TypedDict(
        '_JwtData', {
            'header': _JwtHeader,
            'body': _LaunchData,
        },
        total=False
    )


REQ = t.TypeVar('REQ', bound='Request')
TCONF = t.TypeVar('TCONF', bound='ToolConfAbstract')
SES = t.TypeVar('SES', bound='SessionService')
COOK = t.TypeVar('COOK', bound='CookieService')
T_SELF = t.TypeVar('T_SELF', bound='MessageLaunch')


class MessageLaunch(t.Generic[REQ, TCONF, SES, COOK]):
    __metaclass__ = ABCMeta
    _request = None  # type: REQ
    _tool_config = None  # type: TCONF
    _session_service = None  # type: SES
    _cookie_service = None  # type: COOK
    _jwt = None  # type: _JwtData
    _jwt_verify_options = None  # type: t.Dict[str, bool]
    _registration = None  # type: t.Optional[Registration]
    _launch_id = None  # type: str
    _validated = False  # type: bool
    _auto_validation = True  # type: bool
    _restored = False  # type: bool
    _id_token_hash = None  # type: t.Optional[str]
    _public_key_cache_data_storage = None  # type: t.Optional[LaunchDataStorage[t.Any]]
    _public_key_cache_lifetime = None  # type: t.Optional[int]

    def __init__(self, request, tool_config, session_service, cookie_service, launch_data_storage=None):
        # type: (REQ, TCONF, SES, COOK, t.Optional[LaunchDataStorage[t.Any]]) -> None
        self._request = request
        self._tool_config = tool_config
        self._session_service = session_service
        self._cookie_service = cookie_service
        self._launch_id = "lti1p3-launch-" + str(uuid.uuid4())
        self._jwt = {}
        self._jwt_verify_options = {'verify_aud': False}
        self._id_token_hash = None
        self._validated = False
        self._auto_validation = True
        self._restored = False
        self._public_key_cache_data_storage = None
        self._public_key_cache_lifetime = None

        if launch_data_storage:
            self.set_launch_data_storage(launch_data_storage)

    @abstractmethod
    def _get_request_param(self, key):
        # type: (str) -> t.Any
        raise NotImplementedError

    def set_launch_id(self, launch_id):
        # type: (T_SELF, str) -> T_SELF
        self._launch_id = launch_id
        return self

    def set_auto_validation(self, enable):
        # type: (T_SELF, bool) -> T_SELF
        self._auto_validation = enable
        return self

    def set_jwt(self, val):
        # type: (T_SELF, _JwtData) -> T_SELF
        self._jwt = val
        return self

    def set_jwt_verify_options(self, val):
        # type: (T_SELF, t.Dict[str, bool]) -> T_SELF
        self._jwt_verify_options = val
        return self

    def set_restored(self):
        # type: (T_SELF) -> T_SELF
        self._restored = True
        return self

    def get_session_service(self):
        # type: () -> SES
        return self._session_service

    def get_iss(self):
        # type: () -> str
        iss = self._get_jwt_body().get('iss')
        if not iss:
            raise LtiException('"iss" is empty')
        return iss

    def get_client_id(self):
        # type: () -> str
        jwt_body = self._get_jwt_body()
        aud = jwt_body.get('aud')
        return aud[0] if isinstance(aud, list) else aud  # type: ignore

    @classmethod
    def from_cache(cls, launch_id, request, tool_config, session_service=None, cookie_service=None,
                   launch_data_storage=None):
        # type: (t.Type[T_SELF], str, REQ, TCONF, SES, COOK, t.Optional[LaunchDataStorage[t.Any]]) -> T_SELF
        obj = cls(request, tool_config, session_service=session_service, cookie_service=cookie_service,
                  launch_data_storage=launch_data_storage)
        launch_data = obj.get_session_service().get_launch_data(launch_id)
        if not launch_data:
            raise LtiException("Launch data not found")
        return obj.set_launch_id(launch_id)\
            .set_auto_validation(enable=False)\
            .set_jwt({'body': launch_data})\
            .set_restored()\
            .validate_registration()

    def validate(self):
        # type: (T_SELF) -> T_SELF
        """
        Validates all aspects of an incoming LTI message launch and caches the launch if successful.
        """
        if self._restored:
            raise LtiException("Can't validate restored launch")
        self._validated = True
        try:
            return self.validate_state()\
                .validate_jwt_format()\
                .validate_nonce()\
                .validate_registration()\
                .validate_jwt_signature()\
                .validate_deployment()\
                .validate_message()\
                .save_launch_data()
        except Exception:
            self._validated = False
            raise

    def _get_jwt_body(self):
        # type: () -> _LaunchData
        if not self._validated and self._auto_validation:
            self.validate()
        return self._jwt.get('body', {})

    def _get_iss(self):
        # type: () -> str
        iss = self._get_jwt_body().get('iss')
        if not iss:
            raise LtiException('"iss" is empty')
        return iss

    def _get_id_token(self):
        # type: () -> str
        id_token = self._get_request_param('id_token')
        if not id_token:
            raise LtiException("Missing id_token")
        return id_token

    def _get_id_token_hash(self):
        # type: () -> str
        if not self._id_token_hash:
            id_token = self._get_id_token()
            id_token_param = encode_on_py3(id_token, 'utf8')
            self._id_token_hash = hashlib.md5(id_token_param).hexdigest()
        return self._id_token_hash

    def _get_deployment_id(self):
        # type: () -> str
        deployment_id = self._get_jwt_body().get('https://purl.imsglobal.org/spec/lti/claim/deployment_id')
        if not deployment_id:
            raise LtiException("deployment_id is not set in jwt body")
        return deployment_id

    def has_nrps(self):
        # type: () -> bool
        """
        Returns whether or not the current launch can use the names and roles service.

        :return: bool  Returns a boolean indicating the availability of names and roles.
        """
        return self._get_jwt_body().get('https://purl.imsglobal.org/spec/lti-nrps/claim/namesroleservice', {})\
            .get('context_memberships_url', None) is not None

    def get_nrps(self):
        # type: () -> NamesRolesProvisioningService
        """
        Fetches an instance of the names and roles service for the current launch.

        :return: NamesRolesProvisioningService
        """
        assert self._registration is not None, 'Registration not yet set'
        connector = ServiceConnector(self._registration)
        names_role_service = self._get_jwt_body()\
            .get('https://purl.imsglobal.org/spec/lti-nrps/claim/namesroleservice')
        if not names_role_service:
            raise LtiException('namesroleservice is not set in jwt body')
        return NamesRolesProvisioningService(connector, names_role_service)

    def has_ags(self):
        # type: () -> bool
        """
        Returns whether or not the current launch can use the assignments and grades service.

        :return: bool  Returns a boolean indicating the availability of assignments and grades.
        """
        return self._get_jwt_body().get('https://purl.imsglobal.org/spec/lti-ags/claim/endpoint', None) is not None

    def get_ags(self):
        # type: () -> AssignmentsGradesService
        """
        Fetches an instance of the assignments and grades service for the current launch.

        :return: AssignmentsGradesService
        """
        assert self._registration is not None, 'Registration not yet set'
        connector = ServiceConnector(self._registration)
        endpoint = self._get_jwt_body() \
            .get('https://purl.imsglobal.org/spec/lti-ags/claim/endpoint')
        if not endpoint:
            raise LtiException('endpoint is not set in jwt body')
        return AssignmentsGradesService(connector, endpoint)

    def get_deep_link(self):
        # type: () -> DeepLink
        """
        Fetches a deep link that can be used to construct a deep linking response.

        :return: DeepLink
        """
        assert self._registration is not None, 'Registration not yet set'

        deployment_id = self._get_deployment_id()
        deep_linking_settings = self._get_jwt_body() \
            .get('https://purl.imsglobal.org/spec/lti-dl/claim/deep_linking_settings')
        if not deep_linking_settings:
            raise LtiException('deep_linking_settings is not set in jwt body')

        return DeepLink(self._registration, deployment_id, deep_linking_settings)

    def is_deep_link_launch(self):
        # type: () -> bool
        """
        Returns whether or not the current launch is a deep linking launch.

        :return: bool  Returns true if the current launch is a deep linking launch.
        """
        return self._get_jwt_body() \
                   .get('https://purl.imsglobal.org/spec/lti/claim/message_type', None) == 'LtiDeepLinkingRequest'

    def is_resource_launch(self):
        # type: () -> bool
        """
        Returns whether or not the current launch is a resource launch.

        :return: bool  Returns true if the current launch is a resource launch.
        """
        return self._get_jwt_body() \
                   .get('https://purl.imsglobal.org/spec/lti/claim/message_type', None) == 'LtiResourceLinkRequest'

    def get_launch_data(self):
        # type: () -> _LaunchData
        """
        Fetches the decoded body of the JWT used in the current launch.

        :return: dict  Returns the decoded json body of the launch
        """
        return self._get_jwt_body()

    def get_launch_id(self):
        # type: () -> str
        """
        Get the unique launch id for the current launch.

        :return: str  A unique identifier used to re-reference the current launch in subsequent requests.
        """
        return self._launch_id

    def get_tool_conf(self):
        # type: () -> TCONF
        return self._tool_config

    def urlsafe_b64decode(self, val):
        # type: (str) -> str
        remainder = len(val) % 4
        if remainder > 0:
            padlen = 4 - remainder
            val = val + ('=' * padlen)
        if hasattr(str, 'maketrans'):
            tmp = val.translate(str.maketrans('-_', '+/'))  # type: ignore
            return base64.b64decode(tmp).decode("utf-8")  # type: ignore
        else:
            tmp = str(val).translate(string.maketrans('-_', '+/'))  # type: ignore
            return base64.b64decode(tmp)  # type: ignore

    def set_public_key_caching(self, data_storage, cache_lifetime=7200):
        # type: (LaunchDataStorage[t.Any], int) -> None
        self._public_key_cache_data_storage = data_storage
        self._public_key_cache_lifetime = cache_lifetime

    def fetch_public_key(self, key_set_url):
        # type: (str) -> _KeySet
        cache_key = 'key-set-url-' + hashlib.md5(encode_on_py3(key_set_url, 'utf-8')).hexdigest()

        with DisableSessionId(self._public_key_cache_data_storage):
            if self._public_key_cache_data_storage:
                public_key = self._public_key_cache_data_storage.get_value(cache_key)
                if public_key:
                    return public_key

            try:
                resp = requests.get(key_set_url)
            except requests.exceptions.RequestException as e:
                raise LtiException("Error during fetch URL " + key_set_url + ": " + str(e))
            try:
                public_key = resp.json()
                if self._public_key_cache_data_storage:
                    self._public_key_cache_data_storage.set_value(
                        cache_key, public_key, self._public_key_cache_lifetime)
                return public_key
            except ValueError:
                raise LtiException("Invalid response from " + key_set_url + ". Must be JSON: " + resp.text)

    def get_public_key(self):
        # type: () -> str
        assert self._registration is not None, 'Registration not yet set'
        public_key_set = self._registration.get_key_set()
        key_set_url = self._registration.get_key_set_url()

        if not public_key_set:
            assert key_set_url is not None, 'If public_key_set is not set, public_set_url should be set'
            if key_set_url.startswith(('http://', 'https://')):
                public_key_set = self.fetch_public_key(key_set_url)
                self._registration.set_key_set(public_key_set)
            else:
                raise LtiException("Invalid URL: " + key_set_url)

        # Find key used to sign the JWT (matches the KID in the header)
        kid = self._jwt.get('header', {}).get('kid', None)
        alg = self._jwt.get('header', {}).get('alg', None)

        if not kid:
            raise LtiException("JWT KID not found")
        if not alg:
            raise LtiException("JWT ALG not found")

        for key in public_key_set['keys']:
            key_kid = key.get('kid')
            key_alg = key.get('alg', 'RS256')
            if key_kid and key_kid == kid and key_alg == alg:
                try:
                    key_json = json.dumps(key)
                    jwk_obj = JWK.from_json(key_json)
                    return jwk_obj.export_to_pem()
                except (ValueError, TypeError):
                    raise LtiException("Can't convert JWT key to PEM format")

        # Could not find public key with a matching kid and alg.
        raise LtiException("Unable to find public key")

    def validate_state(self):
        # type: (T_SELF) -> T_SELF
        # Check State for OIDC.
        state_from_request = self._get_request_param('state')
        if not state_from_request:
            raise LtiException("Missing state param")

        id_token_hash = self._get_id_token_hash()
        if not self._session_service.check_state_is_valid(state_from_request, id_token_hash):
            state_from_cookie = self._cookie_service.get_cookie(state_from_request)
            if state_from_request != state_from_cookie:
                # Error if state doesn't match.
                raise LtiException("State not found")

        return self

    def validate_jwt_format(self):
        # type: (T_SELF) -> T_SELF
        id_token = self._get_id_token()
        jwt_parts = id_token.split('.')

        if len(jwt_parts) != 3:
            # Invalid number of parts in JWT.
            raise LtiException("Invalid id_token, JWT must contain 3 parts")

        try:
            # Decode JWT headers.
            header = self.urlsafe_b64decode(jwt_parts[0])
            self._jwt['header'] = json.loads(header)

            # Decode JWT body.
            body = self.urlsafe_b64decode(jwt_parts[1])
            self._jwt['body'] = json.loads(body)
        except Exception:
            raise LtiException("Invalid JWT format, can't be decoded")

        return self

    def validate_nonce(self):
        # type: (T_SELF) -> T_SELF
        nonce = self._get_jwt_body().get('nonce')
        if not nonce:
            raise LtiException('"nonce" is empty')

        res = self._session_service.check_nonce(nonce)
        if not res:
            raise LtiException("Invalid Nonce")

        return self

    def validate_registration(self):
        # type: (T_SELF) -> T_SELF
        iss = self.get_iss()
        jwt_body = self._get_jwt_body()
        client_id = self.get_client_id()

        # Mypy doesn't support higher kinded types yet so it thinks that all
        # generic attrs have type `Any`. See issue:
        # https://github.com/python/mypy/issues/8228
        config = self._tool_config  # type: ToolConfAbstract[REQ]
        req = self._request  # type: REQ

        # Find registration
        if config.check_iss_has_one_client(iss):
            self._registration = config.find_registration(
                iss, action=Action.MESSAGE_LAUNCH, request=req, jwt_body=jwt_body)
        else:
            self._registration = config.find_registration_by_params(
                iss, client_id, action=Action.MESSAGE_LAUNCH, request=req, jwt_body=jwt_body)

        if not self._registration:
            raise LtiException('Registration not found.')

        # Check client id
        if client_id != self._registration.get_client_id():
            raise LtiException("Client id not registered for this issuer")

        return self

    def validate_jwt_signature(self):
        # type: (T_SELF) -> T_SELF
        id_token = self._get_id_token()

        # Fetch public key.
        public_key = self.get_public_key()

        try:
            jwt.decode(id_token, public_key, algorithms=['RS256'], options=self._jwt_verify_options)
        except jwt.InvalidTokenError as e:
            raise LtiException("Can't decode id_token: " + str(e))

        return self

    def validate_deployment(self):
        # type: (T_SELF) -> T_SELF
        iss = self.get_iss()
        client_id = self.get_client_id()
        deployment_id = self._get_deployment_id()
        tool_config = self._tool_config  # type: ToolConfAbstract

        # Find deployment.
        if tool_config.check_iss_has_one_client(iss):
            deployment = tool_config.find_deployment(iss, deployment_id)
        else:
            deployment = tool_config.find_deployment_by_params(iss, deployment_id, client_id)
        if not deployment:
            raise LtiException("Unable to find deployment")

        return self

    def validate_message(self):
        # type: (T_SELF) -> T_SELF
        jwt_body = self._get_jwt_body()
        message_type = jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/message_type', None)
        if not message_type:
            raise LtiException("Invalid message type")

        validators = get_validators()
        validated = False
        for validator in validators:
            if validator.can_validate(jwt_body):
                if validated:
                    raise LtiException("Validator conflict")
                validated = True
                res = validator.validate(jwt_body)
                if not res:
                    raise LtiException("Message validation failed")

        if not validated:
            raise LtiException("Unrecognized message type")

        return self

    def set_launch_data_storage(self, data_storage):
        data_storage.set_request(self._request)
        session_cookie_name = data_storage.get_session_cookie_name()
        if session_cookie_name:
            session_id = self._cookie_service.get_cookie(session_cookie_name)
            if session_id:
                data_storage.set_session_id(session_id)
            else:
                raise LtiException("Missing %s cookie" % session_cookie_name)
        self._session_service.set_data_storage(data_storage)
        return self

    def set_launch_data_lifetime(self, time_sec):
        # type: (T_SELF, int) -> T_SELF
        self._session_service.set_launch_data_lifetime(time_sec)
        return self

    def save_launch_data(self):
        # type: (T_SELF) -> T_SELF
        state_from_request = self._get_request_param('state')
        id_token_hash = self._get_id_token_hash()

        self._session_service.save_launch_data(self._launch_id, self._jwt['body'])
        self._session_service.set_state_valid(state_from_request, id_token_hash)
        return self

    def get_params_from_login(self):
        # type: () -> object
        state = self._get_request_param('state')
        return self._session_service.get_state_params(state)

    def check_jwt_body_is_empty(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return not jwt_body

    def check_staff_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return StaffRole(jwt_body).check()

    def check_student_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return StudentRole(jwt_body).check()

    def check_teacher_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return TeacherRole(jwt_body).check()

    def check_teaching_assistant_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return TeachingAssistantRole(jwt_body).check()

    def check_designer_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return DesignerRole(jwt_body).check()

    def check_observer_access(self):
        # type: () -> bool
        jwt_body = self._get_jwt_body()
        return ObserverRole(jwt_body).check()



from ..exception import LtiException
from .abstract import MessageValidatorAbstract


class ResourceMessageValidator(MessageValidatorAbstract):

    def validate(self, jwt_body):
        self.run_common_validators(jwt_body)

        id_val = jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/resource_link', {}).get('id')
        if not id_val:
            raise LtiException('Missing Resource Link Id')

        return True

    def can_validate(self, jwt_body):
        return jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/message_type') == 'LtiResourceLinkRequest'



import inspect
import sys
import typing as t
import warnings
from abc import ABCMeta, abstractmethod

REQ = t.TypeVar('REQ', bound='Request')

if t.TYPE_CHECKING:
    from ..request import Request
    from ..registration import Registration
    from ..deployment import Deployment
    from ..message_launch import _LaunchData
    from typing_extensions import Literal, Final
    FIND_REG_KWARGS = t.Union[Literal['oidc_login', 'message_launch'], REQ,
                              _LaunchData]
    PossibleRelationTypes = Literal['one-issuer-one-client-id', 'one-issuer-many-client-ids']


class IssuerToClientRelation(object):
    ONE_CLIENT_ID_PER_ISSUER = 'one-issuer-one-client-id'  # type: Final
    MANY_CLIENTS_IDS_PER_ISSUER = 'one-issuer-many-client-ids'  # type: Final


class ToolConfAbstract(t.Generic[REQ]):
    __metaclass__ = ABCMeta
    reg_extended_search = False  # type: bool
    issuers_relation_types = {}  # type: t.MutableMapping[str, PossibleRelationTypes]

    def __init__(self):
        # type: () -> None
        if sys.version_info[0] > 2:
            argspec = inspect.getfullargspec(self.find_registration_by_issuer)
            self.reg_extended_search = None not in (argspec.varargs, argspec.varkw)
        else:
            argspec = inspect.getargspec(self.find_registration_by_issuer)  # pylint: disable=deprecated-method
            self.reg_extended_search = None not in (argspec.varargs, argspec.keywords)

    def check_iss_has_one_client(self, iss):
        # type: (str) -> bool
        """
        Two methods check_iss_has_one_client / check_iss_has_many_clients are needed for the the backward compatibility
        with the previous versions of the library (1.4.0 and early) where ToolConfDict supported only client_id per iss.
        Should return False for all new ToolConf-s
        """
        iss_type = self.issuers_relation_types.get(iss, IssuerToClientRelation.ONE_CLIENT_ID_PER_ISSUER)
        return iss_type == IssuerToClientRelation.ONE_CLIENT_ID_PER_ISSUER

    def check_iss_has_many_clients(self, iss):
        # type: (str) -> bool
        """
        Should return True for all new ToolConf-s
        """
        iss_type = self.issuers_relation_types.get(iss, IssuerToClientRelation.ONE_CLIENT_ID_PER_ISSUER)
        return iss_type == IssuerToClientRelation.MANY_CLIENTS_IDS_PER_ISSUER

    def set_iss_has_one_client(self, iss):
        # type: (str) -> None
        self.issuers_relation_types[iss] = IssuerToClientRelation.ONE_CLIENT_ID_PER_ISSUER

    def set_iss_has_many_clients(self, iss):
        # type: (str) -> None
        self.issuers_relation_types[iss] = IssuerToClientRelation.MANY_CLIENTS_IDS_PER_ISSUER

    def find_registration(self, iss, *args, **kwargs):
        # type: (str, *None, **FIND_REG_KWARGS) -> t.Optional[Registration]
        if self.reg_extended_search:
            return self.find_registration_by_issuer(iss, *args, **kwargs)
        else:
            warnings.warn(
                "Signature of ToolConfAbstract.find_registration_by_issuer method was changed, "
                "please update you custom implementation", DeprecationWarning)
            return self.find_registration_by_issuer(iss)

    @abstractmethod
    def find_registration_by_issuer(self, iss, *args, **kwargs):
        # type: (str, *None, **FIND_REG_KWARGS) -> t.Optional[Registration]
        """
        Find registration in case if iss has only one client id, i.e
        in case of { ... "iss": { ... "client_id: "client" ... }, ... } config.

        You may skip implementation of this method in case if all iss in your config could have more than one client id.
        """
        raise NotImplementedError

    @abstractmethod
    def find_registration_by_params(self, iss, client_id, *args, **kwargs):
        # type: (str, str, *None, **FIND_REG_KWARGS) -> t.Optional[Registration]
        """
        Find registration in case if iss has many client ids, i.e
        in case of { ... "iss": [ { ... "client_id: "client1" ... }, { ... "client_id: "client2" ... } ], ... } config.

        You may skip implementation of this method in case if all iss in your config couldn't have more than one
        client id, but it is outdated and not recommended way of storing configuration.
        """
        raise NotImplementedError

    @abstractmethod
    def find_deployment(self, iss, deployment_id):
        # type: (str, str) -> t.Optional[Deployment]
        """
        Find deployment in case if iss has only one client id, i.e
        in case of { ... "iss": { ... "client_id: "client" ... }, ... } config.

        You may skip implementation of this method in case if all iss in your config could have more than one client id.
        """
        raise NotImplementedError

    @abstractmethod
    def find_deployment_by_params(self, iss, deployment_id, client_id, *args, **kwargs):
        # type: (str, str, str, *None, **None) -> t.Optional[Deployment]
        """
        Find deployment in case if iss has many client ids, i.e
        in case of { ... "iss": [ { ... "client_id: "client1" ... }, { ... "client_id: "client2" ... } ], ... } config.

        You may skip implementation of this method in case if all iss in your config couldn't have more than one
        client id, but it is outdated and not recommended way of storing configuration.
        """
        raise NotImplementedError

    def get_jwks(self, iss=None, client_id=None, **kwargs):
        keys = []
        if iss:
            if self.check_iss_has_one_client(iss):
                reg = self.find_registration(iss)
            elif self.check_iss_has_many_clients(iss):
                reg = self.find_registration_by_params(iss, client_id, **kwargs)
            else:
                raise Exception('Invalid issuer relation type')
            keys = reg.get_jwks()
        return {
            'keys': keys
        }



import json
import typing as t

from jwcrypto.jwk import JWK  # type: ignore

T_SELF = t.TypeVar('T_SELF', bound='Registration')


if t.TYPE_CHECKING:
    from mypy_extensions import TypedDict
    _Key = TypedDict('_Key', {'kid': str, 'alg': str}, total=True)
    _KeySet = TypedDict('_KeySet', {'keys': t.List[_Key]}, total=True)
    from cryptography.hazmat.primitives.asymmetric.rsa import (
        RSAPublicKey, RSAPrivateKey
    )
    _POSSIBLE_KEYS = t.Union[str, bytes, RSAPrivateKey, RSAPublicKey]


class Registration(object):
    _issuer = None  # type: t.Optional[str]
    _client_id = None  # type: t.Optional[str]
    _key_set_url = None  # type: t.Optional[str]
    _key_set = None  # type: t.Optional[_KeySet]
    _auth_token_url = None  # type: t.Optional[str]
    _auth_login_url = None  # type: t.Optional[str]
    _tool_private_key = None  # type: t.Optional[_POSSIBLE_KEYS]
    _auth_audience = None
    _tool_public_key = None

    def get_issuer(self):
        # type: () -> t.Optional[str]
        return self._issuer

    def set_issuer(self, issuer):
        # type: (T_SELF, str) -> T_SELF
        self._issuer = issuer
        return self

    def get_client_id(self):
        # type: () -> t.Optional[str]
        return self._client_id

    def set_client_id(self, client_id):
        # type: (T_SELF, str) -> T_SELF
        self._client_id = client_id
        return self

    def get_key_set(self):
        # type: () -> t.Optional[_KeySet]
        return self._key_set

    def set_key_set(self, key_set):
        # type: (T_SELF, t.Optional[_KeySet]) -> T_SELF
        self._key_set = key_set
        return self

    def get_key_set_url(self):
        # type: () -> t.Optional[str]
        return self._key_set_url

    def set_key_set_url(self, key_set_url):
        # type: (T_SELF, str) -> T_SELF
        self._key_set_url = key_set_url
        return self

    def get_auth_token_url(self):
        # type: () -> t.Optional[str]
        return self._auth_token_url

    def set_auth_token_url(self, auth_token_url):
        # type: (T_SELF, str) -> T_SELF
        self._auth_token_url = auth_token_url
        return self

    def get_auth_login_url(self):
        # type: () -> t.Optional[str]
        return self._auth_login_url

    def set_auth_login_url(self, auth_login_url):
        # type: (T_SELF, str) -> T_SELF
        self._auth_login_url = auth_login_url
        return self

    def get_auth_audience(self):
        return self._auth_audience

    def set_auth_audience(self, auth_audience):
        self._auth_audience = auth_audience
        return self

    def get_tool_private_key(self):
        # type: () -> t.Optional[_POSSIBLE_KEYS]
        return self._tool_private_key

    def set_tool_private_key(self, tool_private_key):
        # type: (T_SELF, _POSSIBLE_KEYS) -> T_SELF
        self._tool_private_key = tool_private_key
        return self

    def get_tool_public_key(self):
        return self._tool_public_key

    def set_tool_public_key(self, tool_public_key):
        self._tool_public_key = tool_public_key
        return self

    @classmethod
    def get_jwk(cls, public_key):
        # type: (str) -> t.Mapping[str, t.Any]
        jwk_obj = JWK.from_pem(public_key.encode('utf-8'))
        public_jwk = json.loads(jwk_obj.export_public())
        public_jwk['alg'] = 'RS256'
        public_jwk['use'] = 'sig'
        return public_jwk

    def get_jwks(self):
        # type: () -> t.List[t.Mapping[str, t.Any]]
        keys = []
        public_key = self.get_tool_public_key()
        if public_key:
            keys.append(Registration.get_jwk(public_key))
        return keys

    def get_kid(self):
        # type: () -> t.Optional[str]
        key = self.get_tool_public_key()
        if key:
            jwk = Registration.get_jwk(key)
            return jwk.get('kid') if jwk else None
        return None



# mypy: ignore-errors
import json

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _
from pylti1p3.registration import Registration


class LtiToolKey(models.Model):
    name = models.CharField(max_length=255, null=False, blank=False, unique=True, help_text=_("Key name"))
    private_key = models.TextField(null=False, blank=False, help_text=_("Tool's generated Private key. "
                                                                        "Keep this value in secret"))
    public_key = models.TextField(null=True, blank=True, help_text=_("Tool's generated Public key"))
    public_jwk = models.TextField(null=True, blank=True, help_text=_("Tool's generated Public key (from the field"
                                                                     " above) presented as JWK."))

    def save(self, *args, **kwargs):  # pylint: disable=arguments-differ,signature-differs
        if self.public_key:
            public_jwk_dict = Registration.get_jwk(self.public_key)
            self.public_jwk = json.dumps(public_jwk_dict)
        else:
            self.public_key = None
            self.public_jwk = None
        super(LtiToolKey, self).save(*args, **kwargs)

    def __str__(self):
        return '<LtiToolKey id=%d, name=%s>' % (self.id, self.name)

    class Meta(object):
        db_table = "lti1p3_tool_key"
        verbose_name = "lti 1.3 tool key"
        verbose_name_plural = "lti 1.3 tool keys"


class LtiTool(models.Model):
    title = models.CharField(max_length=255)
    is_active = models.BooleanField(default=True)
    issuer = models.CharField(max_length=255,
                              help_text=_("This will usually look something like 'http://example.com'. "
                                          "Value provided by LTI 1.3 Platform"))
    client_id = models.CharField(max_length=255, null=False, blank=False,
                                 help_text=_("Value provided by LTI 1.3 Platform"))
    use_by_default = models.BooleanField(default=False, help_text=_("This iss config will be used in case "
                                                                    "if client-id was not passed"))
    auth_login_url = models.CharField(max_length=1024, null=False, blank=False,
                                      help_text=_("The platform's OIDC login endpoint. "
                                                  "Value provided by LTI 1.3 Platform"),
                                      validators=[URLValidator()])
    auth_token_url = models.CharField(max_length=1024, null=False, blank=False,
                                      help_text=_("The platform's service authorization "
                                                  "endpoint. Value provided by "
                                                  "LTI 1.3 Platform"),
                                      validators=[URLValidator()])
    auth_audience = models.CharField(max_length=1024, null=True, blank=True,
                                     help_text=_("The platform's OAuth2 Audience (aud). "
                                                 "Usually could be skipped"))
    key_set_url = models.CharField(max_length=1024, null=True, blank=True,
                                   help_text=_("The platform's JWKS endpoint. "
                                               "Value provided by LTI 1.3 Platform"),
                                   validators=[URLValidator()])
    key_set = models.TextField(null=True, blank=True, help_text=_("In case if platform's JWKS endpoint somehow "
                                                                  "unavailable you may paste JWKS here. "
                                                                  "Value provided by LTI 1.3 Platform"))
    tool_key = models.ForeignKey(LtiToolKey, on_delete=models.PROTECT, related_name="lti_tools")
    deployment_ids = models.TextField(null=False, blank=False,
                                      help_text=_("List of Deployment IDs. "
                                                  "Example: [\"test-id-1\", \"test-id-2\", ...] "
                                                  "Each value is provided by LTI 1.3 Platform. "))

    def clean(self):
        if not self.key_set_url and not self.key_set:
            raise ValidationError({'key_set_url': _('Even one of "key_set_url" or "key_set" should be set')})

        if self.key_set:
            key_set_valid = False
            try:
                key_set_data = json.loads(self.key_set)
                if isinstance(key_set_data, dict):
                    key_set_valid = True
            except ValueError:
                pass
            if not key_set_valid:
                raise ValidationError({'key_set': _('Should be a dict')})

        deployment_ids_valid = False
        try:
            deployment_ids_data = json.loads(self.deployment_ids)
            if isinstance(deployment_ids_data, list):
                deployment_ids_valid = True
        except ValueError:
            pass
        if not deployment_ids_valid:
            raise ValidationError({'deployment_ids': _('Should be a list. Example: ["test-id-1", "test-id-2", ...]')})

    def to_dict(self):
        data = {
            "issuer": self.issuer,
            "client_id": self.client_id,
            "auth_login_url": self.auth_login_url,
            "auth_token_url": self.auth_token_url,
            "auth_audience": self.auth_audience,
            "key_set_url": self.key_set_url,
            "key_set": json.loads(self.key_set) if self.key_set else None,
            "deployment_ids": json.loads(self.deployment_ids) if self.deployment_ids else []
        }
        return data

    class Meta(object):
        unique_together = [
            ['issuer', 'client_id'],
        ]
        db_table = "lti1p3_tool"
        verbose_name = "lti 1.3 tool"
        verbose_name_plural = "lti 1.3 tools"



import sys
import time
import typing as t
import uuid

import jwt  # type: ignore

if t.TYPE_CHECKING:
    from .deep_link_resource import DeepLinkResource
    from .registration import Registration
    from typing_extensions import Literal
    from mypy_extensions import TypedDict

    _DeepLinkData = TypedDict(
        '_DeepLinkData',
        {
            # Required data:
            'deep_link_return_url': str,
            'accept_types': t.List[Literal['link', 'ltiResourceLink']],
            'accept_presentation_document_targets': t.List[
                Literal['iframe', 'window', 'embed']],

            # Optional data
            'accept_multiple': t.Union[bool, Literal['true', 'false']],
            'auto_create': t.Union[bool, Literal['true', 'false']],
            'title': str,
            'text': str,
            'data': object,
        },
        total=False,
    )


class DeepLink(object):
    _registration = None  # type: Registration
    _deployment_id = None  # type: str
    _deep_link_settings = None  # type: _DeepLinkData

    def __init__(self, registration, deployment_id, deep_link_settings):
        # type: (Registration, str, _DeepLinkData) -> None
        self._registration = registration
        self._deployment_id = deployment_id
        self._deep_link_settings = deep_link_settings

    def _generate_nonce(self):
        return uuid.uuid4().hex + uuid.uuid1().hex

    def get_message_jwt(self, resources):
        # type: (t.Sequence[DeepLinkResource]) -> t.Dict[str, object]
        message_jwt = {
            'iss': self._registration.get_client_id(),
            'aud': [self._registration.get_issuer()],
            'exp': int(time.time()) + 600,
            'iat': int(time.time()),
            'nonce': 'nonce-' + self._generate_nonce(),
            'https://purl.imsglobal.org/spec/lti/claim/deployment_id': self._deployment_id,
            'https://purl.imsglobal.org/spec/lti/claim/message_type': 'LtiDeepLinkingResponse',
            'https://purl.imsglobal.org/spec/lti/claim/version': '1.3.0',
            'https://purl.imsglobal.org/spec/lti-dl/claim/content_items': [r.to_dict() for r in resources],
            'https://purl.imsglobal.org/spec/lti-dl/claim/data': self._deep_link_settings.get('data')
        }
        return message_jwt

    def encode_jwt(self, message):
        headers = None
        kid = self._registration.get_kid()
        if kid:
            headers = {'kid': kid}
        encoded_jwt = jwt.encode(message, self._registration.get_tool_private_key(), algorithm='RS256',
                                 headers=headers)
        return encoded_jwt.decode('utf-8') if sys.version_info[0] > 2 else encoded_jwt

    def get_response_jwt(self, resources):
        # type: (t.Sequence[DeepLinkResource]) -> str
        message_jwt = self.get_message_jwt(resources)
        return self.encode_jwt(message_jwt)

    def get_response_form_html(self, jwt_val):
        # type: (str) -> str
        html = '<form id="lti13_deep_link_auto_submit" action="%s" method="POST">' \
               '<input type="hidden" name="JWT" value="%s" /></form>' \
               '<script type="text/javascript">document.getElementById(\'lti13_deep_link_auto_submit\').submit();' \
               '</script>' % (self._deep_link_settings['deep_link_return_url'], jwt_val)
        return html

    def output_response_form(self, resources):
        # type: (t.List[DeepLinkResource]) -> str
        jwt_val = self.get_response_jwt(resources)
        return self.get_response_form_html(jwt_val)



from ..exception import LtiException
from .abstract import MessageValidatorAbstract


class DeepLinkMessageValidator(MessageValidatorAbstract):

    def validate(self, jwt_body):
        self.run_common_validators(jwt_body)

        if not jwt_body.get('https://purl.imsglobal.org/spec/lti-dl/claim/deep_linking_settings'):
            raise LtiException('Missing Deep Linking Settings')

        deep_link_settings = jwt_body.get('https://purl.imsglobal.org/spec/lti-dl/claim/deep_linking_settings')
        if not deep_link_settings:
            raise LtiException('Missing Deep Linking Return URL')

        accept_types = deep_link_settings.get('accept_types')

        if not isinstance(accept_types, list) or 'ltiResourceLink' not in accept_types:
            raise LtiException('Must support resource link placement types')

        if not deep_link_settings.get('accept_presentation_document_targets'):
            raise LtiException('Must support a presentation type')

        return True

    def can_validate(self, jwt_body):
        return jwt_body.get('https://purl.imsglobal.org/spec/lti/claim/message_type') == 'LtiDeepLinkingRequest'



import typing as t

if t.TYPE_CHECKING:
    from .lineitem import LineItem
    T_SELF = t.TypeVar('T_SELF', bound='DeepLinkResource')


class DeepLinkResource(object):
    _type = 'ltiResourceLink'  # type: str
    _title = None  # type: t.Optional[str]
    _url = None  # type: t.Optional[str]
    _lineitem = None  # type: t.Optional[LineItem]
    _custom_params = {}  # type: t.Mapping[str, str]
    _target = 'iframe'  # type: str

    def get_type(self):
        # type: () -> str
        return self._type

    def set_type(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._type = value
        return self

    def get_title(self):
        # type: () -> t.Optional[str]
        return self._title

    def set_title(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._title = value
        return self

    def get_url(self):
        # type: () -> t.Optional[str]
        return self._url

    def set_url(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._url = value
        return self

    def get_lineitem(self):
        # type: () -> t.Optional[LineItem]
        return self._lineitem

    def set_lineitem(self, value):
        # type: (T_SELF, LineItem) -> T_SELF
        self._lineitem = value
        return self

    def get_custom_params(self):
        # type: () -> t.Mapping[str, str]
        return self._custom_params

    def set_custom_params(self, value):
        # type: (T_SELF, t.Mapping[str, str]) -> T_SELF
        self._custom_params = value
        return self

    def get_target(self):
        # type: () -> str
        return self._target

    def set_target(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._target = value
        return self

    def to_dict(self):
        # type: () -> t.Dict[str, object]
        res = {
            'type': self._type,
            'title': self._title,
            'url': self._url,
            'custom': self._custom_params
        }  # type: t.Dict[str, object]
        if self._lineitem:
            line_item = {
                'scoreMaximum': self._lineitem.get_score_maximum(),
            }  # type: t.Dict[str, object]

            label = self._lineitem.get_label()
            if label:
                line_item['label'] = label

            resource_id = self._lineitem.get_resource_id()
            if resource_id:
                line_item['resourceId'] = resource_id

            tag = self._lineitem.get_tag()
            if tag:
                line_item['tag'] = tag

            res['lineItem'] = line_item
        return res



from flask import make_response, redirect

from pylti1p3.redirect import Redirect


class FlaskRedirect(Redirect):
    _location = None
    _cookie_service = None

    def __init__(self, location, cookie_service=None):
        super(FlaskRedirect, self).__init__()
        self._location = location
        self._cookie_service = cookie_service

    def do_redirect(self):
        return self._process_response(redirect(self._location))

    def do_js_redirect(self):
        return self._process_response(
            make_response('<script type="text/javascript">window.location="{}";'
                          '</script>'.format(self._location))
        )

    def set_redirect_url(self, location):
        self._location = location

    def get_redirect_url(self):
        return self._location

    def _process_response(self, response):
        if self._cookie_service:
            self._cookie_service.update_response(response)
        return response



import typing as t
import uuid
from abc import ABCMeta, abstractmethod

from .actions import Action
from .cookies_allowed_check import CookiesAllowedCheckPage
from .exception import OIDCException

if t.TYPE_CHECKING:
    from .cookie import CookieService
    from .request import Request
    from .session import SessionService
    from .redirect import Redirect
    from .tool_config import ToolConfAbstract
    from .registration import Registration
    from .launch_data_storage.base import LaunchDataStorage

    def urlencode(url):
        # type: (t.Dict[str, str]) -> str
        return str(url)
else:
    try:
        from urllib import urlencode
    except ImportError:
        from urllib.parse import urlencode

RED = t.TypeVar('RED')
REQ = t.TypeVar('REQ', bound='Request')
TCONF = t.TypeVar('TCONF', bound='ToolConfAbstract')
SES = t.TypeVar('SES', bound='SessionService')
COOK = t.TypeVar('COOK', bound='CookieService')

T_SELF = t.TypeVar('T_SELF', bound='OIDCLogin')


class OIDCLogin(t.Generic[REQ, TCONF, SES, COOK, RED]):
    __metaclass__ = ABCMeta
    _request = None  # type: REQ
    _tool_config = None  # type: TCONF
    _session_service = None  # type: SES
    _cookie_service = None  # type: COOK
    _launch_data_storage = None  # type: t.Optional[LaunchDataStorage[t.Any]]
    _registration = None  # type: Registration

    _cookies_check = False  # type: bool
    _cookies_check_loading_text = 'Loading...'  # type: str
    _cookies_unavailable_msg_main_text = 'Your browser prohibits to save cookies in the iframes.'  # type: str
    _cookies_unavailable_msg_click_text = 'Click here to open content in the new tab.'  # type: str
    _state_params = {}  # type: t.Dict[str, object]

    def __init__(self, request, tool_config, session_service, cookie_service, launch_data_storage=None):
        # type: (REQ, TCONF, SES, COOK, t.Optional[LaunchDataStorage[t.Any]]) -> None
        self._request = request
        self._tool_config = tool_config
        self._session_service = session_service
        self._cookie_service = cookie_service
        self._launch_data_storage = launch_data_storage

    @abstractmethod
    def get_redirect(self, url):
        # type: (str) -> Redirect[RED]
        raise NotImplementedError

    def get_response(self, html):  # pylint: disable=unused-argument
        # type: (str) -> RED
        return ''  # type: ignore

    def get_iss(self):
        # type: () -> t.Optional[str]
        if self._registration:
            return self._registration.get_issuer()
        return None

    def get_client_id(self):
        if self._registration:
            return self._registration.get_client_id()
        return None

    def _get_request_param(self, key):
        # type: (str) -> t.Any
        return self._request.get_param(key)

    def _get_uuid(self):
        # type: () -> str
        return str(uuid.uuid4())

    def _generate_nonce(self):
        # type: () -> str
        return uuid.uuid4().hex + uuid.uuid1().hex

    def _is_new_window_request(self):
        # type: () -> bool
        lti_new_window = self._get_request_param('lti1p3_new_window')
        return bool(lti_new_window)

    def _prepare_redirect_url(self, launch_url):
        # type: (str) -> str
        if not launch_url:
            raise OIDCException("No launch URL configured")

        if self._launch_data_storage:
            self.set_launch_data_storage(self._launch_data_storage)

        # validate request
        self._registration = self.validate_oidc_login()

        # build OIDC Auth Response

        # generate state
        # set cookie (short lived)
        state = 'state-' + self._get_uuid()
        self._cookie_service.set_cookie(state, state, 5 * 60)  # 5 min

        # generate nonce
        nonce = self._generate_nonce()
        self._session_service.save_nonce(nonce)
        if self._state_params:
            self._session_service.save_state_params(state, self._state_params)

        # build Response
        client_id = self._registration.get_client_id()  # Registered client id
        assert client_id is not None, 'Client id should not be None'
        auth_login_url = self._registration.get_auth_login_url()
        assert auth_login_url is not None, 'Auth login url should not be None'

        auth_params = {
            'scope': 'openid',  # OIDC Scope
            'response_type': 'id_token',  # OIDC response is always an id token
            'response_mode': 'form_post',  # OIDC response is always a form post
            'prompt': 'none',  # Don't prompt user on redirect
            'client_id': client_id,  # Registered client id
            'redirect_uri': launch_url,  # URL to return to after login
            'state': state,  # State to identify browser session
            'nonce': nonce,  # Prevent replay attacks
            'login_hint':  self._get_request_param('login_hint')  # Login hint to identify platform session
        }

        # pass back LTI message hint if we have it
        lti_message_hint = self._get_request_param('lti_message_hint')
        if lti_message_hint:
            # LTI message hint to identify LTI context within the platform
            auth_params['lti_message_hint'] = lti_message_hint

        auth_login_return_url = auth_login_url + "?" + urlencode(auth_params)
        return auth_login_return_url

    def _prepare_redirect(self, launch_url):
        # type: (str) -> Redirect[RED]
        auth_login_return_url = self._prepare_redirect_url(launch_url)
        return self.get_redirect(auth_login_return_url)

    def redirect(self, launch_url, js_redirect=False):
        # type: (str, bool) -> RED
        """
        Calculate the redirect location to return to based on an OIDC third party initiated login request.

        :param launch_url: URL to redirect back to after the OIDC login.
        This URL must match exactly a URL white listed in the platform.
        :param js_redirect: Redirect through JS
        :return: Returns a redirect object containing the fully formed OIDC login URL.
        """
        if self._cookies_check:
            if not self._is_new_window_request():
                html = self.get_cookies_allowed_js_check()
                return self.get_response(html)

        redirect_obj = self._prepare_redirect(launch_url)
        if js_redirect:
            return redirect_obj.do_js_redirect()
        return redirect_obj.do_redirect()

    def get_redirect_object(self, launch_url):
        # type: (str) -> Redirect[RED]
        return self._prepare_redirect(launch_url)

    def validate_oidc_login(self):
        # type: () -> Registration
        # validate Issuer
        iss = self._get_request_param('iss')
        if not iss:
            raise OIDCException('Could not find issuer')

        # validate login hint
        login_hint = self._get_request_param('login_hint')
        if not login_hint:
            raise OIDCException('Could not find login hint')

        client_id = self._get_request_param('client_id')

        # fetch registration details
        if self._tool_config.check_iss_has_one_client(iss):
            registration = self._tool_config.find_registration(iss, action=Action.OIDC_LOGIN, request=self._request)
        else:
            registration = self._tool_config.find_registration_by_params(iss, client_id, action=Action.OIDC_LOGIN,
                                                                         request=self._request)

        # check we got something
        if not registration:
            raise OIDCException("Could not find registration details")

        return registration

    def pass_params_to_launch(self, params):
        # type: (T_SELF, t.Dict[str, object]) -> T_SELF
        """
        Ability to pass custom params from oidc login to launch.
        """
        self._state_params = params
        return self

    def enable_check_cookies(self, main_msg=None, click_msg=None, loading_msg=None, **kwargs):
        # type: (T_SELF, t.Optional[str], t.Optional[str], t.Optional[str], **None) -> T_SELF
        # pylint: disable=unused-argument
        self._cookies_check = True
        if main_msg:
            self._cookies_unavailable_msg_main_text = main_msg
        if click_msg:
            self._cookies_unavailable_msg_click_text = click_msg
        if loading_msg:
            self._cookies_check_loading_text = loading_msg
        return self

    def disable_check_cookies(self):
        # type: (T_SELF) -> T_SELF
        self._cookies_check = False
        return self

    def get_additional_login_params(self):
        # type: () -> t.List[str]
        """
        You may add additional custom params in your own OIDCLogin class
        :return: list
        """
        return []

    def get_cookies_allowed_js_check(self):
        # type: () -> str
        protocol = 'https' if self._request.is_secure() else 'http'
        params_lst = ['iss', 'login_hint', 'target_link_uri', 'lti_message_hint',
                      'lti_deployment_id', 'client_id']
        additional_login_params = self.get_additional_login_params()
        params_lst.extend(additional_login_params)

        params = {
            'lti1p3_new_window': '1'
        }
        for param_key in params_lst:
            param_value = self._get_request_param(param_key)
            if param_value:
                params[param_key] = param_value

        page = CookiesAllowedCheckPage(params, protocol, self._cookies_unavailable_msg_main_text,
                                       self._cookies_unavailable_msg_click_text, self._cookies_check_loading_text)

        return page.get_html()

    def set_launch_data_storage(self, data_storage):
        # type: (T_SELF, LaunchDataStorage[t.Any]) -> T_SELF
        data_storage.set_request(self._request)
        session_cookie_name = data_storage.get_session_cookie_name()
        if session_cookie_name:
            session_id = self._cookie_service.get_cookie(session_cookie_name)
            if not session_id:
                session_id = self._get_uuid()
                self._cookie_service.set_cookie(session_cookie_name, session_id, None)
            data_storage.set_session_id(session_id)
        self._session_service.set_data_storage(data_storage)
        return self

    def set_launch_data_lifetime(self, time_sec):
        # type: (T_SELF, int) -> T_SELF
        self._session_service.set_launch_data_lifetime(time_sec)
        return self



from pylti1p3.message_launch import MessageLaunch
from pylti1p3.request import Request
from .cookie import DjangoCookieService
from .request import DjangoRequest
from .session import DjangoSessionService


class DjangoMessageLaunch(MessageLaunch):

    def __init__(self, request, tool_config, session_service=None, cookie_service=None, launch_data_storage=None):
        django_request = request if isinstance(request, Request) else DjangoRequest(request, post_only=True)
        cookie_service = cookie_service if cookie_service else DjangoCookieService(django_request)
        session_service = session_service if session_service else DjangoSessionService(request)
        super(DjangoMessageLaunch, self).__init__(django_request, tool_config, session_service, cookie_service,
                                                  launch_data_storage)

    def _get_request_param(self, key):
        return self._request.get_param(key)



import typing as t

from flask import request
from flask import session as flask_session
from pylti1p3.request import Request

if t.TYPE_CHECKING:
    from pylti1p3.request import SessionLike


class FlaskRequest(Request):
    session = None  # type: SessionLike
    _cookies = None
    _request_data = None
    _request_is_secure = None

    def __init__(self, cookies=None, session=None, request_data=None, request_is_secure=None):
        super(FlaskRequest, self).__init__()
        self._cookies = request.cookies if cookies is None else cookies
        self.session = flask_session if session is None else session
        self._request_is_secure = request.is_secure if request_is_secure is None else request_is_secure

        if request_data:
            self._request_data = request_data

    def get_param(self, key):
        if self._request_data:
            return self._request_data.get(key)
        else:
            if request.method == 'GET':
                return request.args.get(key, None)
            else:
                return request.form.get(key, None)

    def get_cookie(self, key):
        return self._cookies.get(key)

    def is_secure(self):
        return self._request_is_secure



from pylti1p3.launch_data_storage.cache import CacheDataStorage


class FlaskCacheDataStorage(CacheDataStorage):
    _cache = None

    def __init__(self, cache, **kwargs):
        self._cache = cache
        super(FlaskCacheDataStorage, self).__init__(cache, **kwargs)



import json
import typing as t


class CookiesAllowedCheckPage(object):
    _params = {}  # type: t.Mapping[str, str]
    _protocol = 'http'  # type: str
    _main_text = ''  # type: str
    _click_text = ''  # type: str
    _loading_text = ''  # type: str

    def __init__(self, params, protocol, main_text, click_text, loading_text, *args, **kwargs):
        # type: (t.Mapping[str, str], str, str, str, str, *None, **None) -> None
        # pylint: disable=unused-argument
        self._params = params
        self._protocol = protocol
        self._main_text = main_text
        self._click_text = click_text
        self._loading_text = loading_text

    def get_css_block(self):
        # type: () -> str
        css_block = """\
        body {
        font-family: Geneva, Arial, Helvetica, sans-serif;
        }
        """
        return css_block

    def get_js_block(self):
        # type: () -> str
        js_block = """\
        var siteProtocol = '%s';
        var urlParams = %s;

        function getUpdatedUrl() {
            var newSearchParams = [];
            for (var key in urlParams) {
                if (window.location.search.indexOf(key + '=') === -1) {
                    newSearchParams.push(key + '=' + encodeURIComponent(urlParams[key]));
                }
            }
            var searchParamsStr = newSearchParams.join('&');
            if (window.location.search !== '') {
                searchParamsStr = window.location.search + '&' + searchParamsStr;
            } else {
                searchParamsStr = '?' + searchParamsStr;
            }
            return window.location.protocol + '//' + window.location.hostname +
                (window.location.port ? (":" + window.location.port) : "") +
                window.location.pathname + searchParamsStr;
        }

        function displayLoadingBlock() {
            document.getElementById("lti1p3-loading-msg").style.display = "block";
        }

        function displayWarningBlock() {
            document.getElementById("lti1p3-warning-msg").style.display = "block";
            var newTabLink = document.getElementById("lti1p3-new-tab-link");
            var contentUrl = getUpdatedUrl();
            newTabLink.onclick = function() {
                window.open(contentUrl , '_blank');
                newTabLink.parentNode.removeChild(newTabLink);
            };
        }

        function checkCookiesAllowed() {
            var cookie = "lti1p3_test_cookie=1; path=/";
            if (siteProtocol === 'https') {
                cookie = cookie + '; SameSite=None; secure';
            }
            document.cookie = cookie;
            var res = document.cookie.indexOf("lti1p3_test_cookie") !== -1;
            if (res) {
                // remove test cookie and reload page
                document.cookie = "lti1p3_test_cookie=1; expires=Thu, 01-Jan-1970 00:00:01 GMT";
                displayLoadingBlock();
                window.location.href = getUpdatedUrl();
            } else {
                displayWarningBlock();
            }
        }

        document.addEventListener("DOMContentLoaded", checkCookiesAllowed);
        """
        js_block = js_block % (self._protocol, json.dumps(self._params))
        return js_block

    def get_header_block(self):
        # type: () -> str
        return ''

    def get_html(self):
        # type: () -> str
        html = """\
        <!DOCTYPE html>
        <html lang="en">
        <head>
        <title></title>
        <meta charset="UTF-8">
        <style type="text/css">
        {css_block}
        </style>
        <script type="text/javascript">
        {js_block}
        </script>
        </head>
        <body>
        <div id="lti1p3-loading-msg" style="display: none;">
        {loading_text}
        </div>
        <div id="lti1p3-warning-msg" style="display: none;">
        {header_block}
        <p><strong>{main_text}</strong> <a href="javascript: void(0);" id="lti1p3-new-tab-link">{click_text}</a></p>
        </div>
        </body>
        </html>
        """
        html = html.format(
            css_block=self.get_css_block(),
            js_block=self.get_js_block(),
            loading_text=self._loading_text,
            header_block=self.get_header_block(),
            main_text=self._main_text,
            click_text=self._click_text)
        return html



from pylti1p3.message_launch import MessageLaunch
from .cookie import FlaskCookieService
from .session import FlaskSessionService


class FlaskMessageLaunch(MessageLaunch):

    def __init__(self, request, tool_config, session_service=None, cookie_service=None, launch_data_storage=None):
        cookie_service = cookie_service if cookie_service else FlaskCookieService(request)
        session_service = session_service if session_service else FlaskSessionService(request)
        super(FlaskMessageLaunch, self).__init__(request, tool_config, session_service, cookie_service,
                                                 launch_data_storage)

    def _get_request_param(self, key):
        return self._request.get_param(key)



from pylti1p3.request import Request


class DjangoRequest(Request):
    _request = None
    _post_only = False
    _default_params = None

    @property
    def session(self):
        return self._request.session

    def __init__(self, request, post_only=False, default_params=None):
        self.set_request(request)
        self._post_only = post_only
        self._default_params = default_params if default_params else {}

    def set_request(self, request):
        self._request = request

    def get_param(self, key):
        if self._post_only:
            return self._request.POST.get(key, self._default_params.get(key))
        return self._request.GET.get(key, self._request.POST.get(key, self._default_params.get(key)))

    def get_cookie(self, key):
        return self._request.COOKIES.get(key)

    def is_secure(self):
        return self._request.is_secure()



import json
import typing as t

if t.TYPE_CHECKING:
    T_SELF = t.TypeVar('T_SELF', bound='LineItem')


class LineItem(object):
    _id = None  # type: t.Optional[str]
    _score_maximum = None  # type: t.Optional[float]
    _label = None  # type: t.Optional[str]
    _resource_id = None  # type: t.Optional[str]
    _tag = None  # type: t.Optional[str]
    _start_date_time = None  # type: t.Optional[str]
    _end_date_time = None  # type: t.Optional[str]

    def __init__(self, lineitem=None):
        # type: (t.Optional[t.Mapping[str, t.Any]]) -> None
        if not lineitem:
            lineitem = {}
        self._id = lineitem.get("id")
        self._score_maximum = lineitem.get("scoreMaximum")
        self._label = lineitem.get("label")
        self._resource_id = lineitem.get("resourceId")
        self._tag = lineitem.get("tag")
        self._start_date_time = lineitem.get("startDateTime")
        self._end_date_time = lineitem.get("endDateTime")

    def get_id(self):
        # type: () -> t.Optional[str]
        return self._id

    def set_id(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._id = value
        return self

    def get_label(self):
        # type: () -> t.Optional[str]
        return self._label

    def set_label(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._label = value
        return self

    def get_score_maximum(self):
        # type: () -> t.Optional[float]
        return self._score_maximum

    def set_score_maximum(self, value):
        # type: (T_SELF, float) -> T_SELF
        self._score_maximum = value
        return self

    def get_resource_id(self):
        # type: () -> t.Optional[str]
        return self._resource_id

    def set_resource_id(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._resource_id = value
        return self

    def get_tag(self):
        # type: () -> t.Optional[str]
        return self._tag

    def set_tag(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._tag = value
        return self

    def get_start_date_time(self):
        # type: () -> t.Optional[str]
        return self._start_date_time

    def set_start_date_time(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._start_date_time = value
        return self

    def get_end_date_time(self):
        # type: () -> t.Optional[str]
        return self._end_date_time

    def set_end_date_time(self, value):
        # type: (T_SELF, str) -> T_SELF
        self._end_date_time = value
        return self

    def get_value(self):
        # type: () -> str
        data = {
            'id': self._id if self._id else None,
            'scoreMaximum': self._score_maximum if self._score_maximum else None,
            'label': self._label,
            'resourceId': self._resource_id,
            'tag': self._tag,
            'startDateTime': self._start_date_time,
            'endDateTime': self._end_date_time
        }
        return json.dumps({k: v for k, v in data.items() if v})



from django.http import HttpResponse  # type: ignore
from pylti1p3.oidc_login import OIDCLogin
from pylti1p3.request import Request

from .cookie import DjangoCookieService
from .redirect import DjangoRedirect
from .request import DjangoRequest
from .session import DjangoSessionService


class DjangoOIDCLogin(OIDCLogin):

    def __init__(self, request, tool_config, session_service=None, cookie_service=None, launch_data_storage=None):
        django_request = request if isinstance(request, Request) else DjangoRequest(request)
        cookie_service = cookie_service if cookie_service else DjangoCookieService(django_request)
        session_service = session_service if session_service else DjangoSessionService(request)
        super(DjangoOIDCLogin, self).__init__(django_request, tool_config, session_service, cookie_service,
                                              launch_data_storage)

    def get_redirect(self, url):
        return DjangoRedirect(url, self._cookie_service)

    def get_response(self, html):
        return HttpResponse(html)



from __future__ import print_function

import sys

from setuptools import find_packages, setup

from pylti1p3 import __version__

if sys.version_info < (2, 7):
    error = "ERROR: PyLTI1p3 requires Python 2.7+ ... exiting."
    print(error, file=sys.stderr)
    sys.exit(1)


install_requires = [
    'pyjwt>=1.5',
    'jwcrypto',
    'requests'
]

if sys.version_info < (3, 5):
    extra_pkgs = ['typing']
else:
    extra_pkgs = []

with open("README.rst", "rt") as readme:
    long_description = readme.read().strip()

packages = find_packages(exclude=["examples", "tests"])

setup(
    name='PyLTI1p3',
    version=__version__,
    description='LTI 1.3 Advantage Tool implementation in Python',
    keywords="pylti,pylti1p3,lti,lti1.3,lti1p3,django,flask",
    author='Dmitry Viskov',
    author_email='dmitry.viskov@webenterprise.ru',
    maintainer="Dmitry Viskov",
    long_description=long_description,
    install_requires=install_requires + extra_pkgs,
    license='MIT',
    url='https://github.com/dmitry-viskov/pylti1.3',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'Environment :: Web Environment',
        "Framework :: Django",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    package_data={
        'pylti1p3': ['py.typed'],
        'pylti1p3.tool_config': ['py.typed'],
        'pylti1p3.contrib': ['py.typed'],
    },
)



import typing as t
from abc import ABCMeta, abstractmethod

T = t.TypeVar('T')


class Redirect(t.Generic[T]):
    __metaclass__ = ABCMeta

    @abstractmethod
    def do_redirect(self):
        # type: () -> T
        raise NotImplementedError

    @abstractmethod
    def do_js_redirect(self):
        # type: () -> T
        raise NotImplementedError

    @abstractmethod
    def set_redirect_url(self, location):
        # type: (str) -> None
        raise NotImplementedError

    @abstractmethod
    def get_redirect_url(self):
        # type: () -> str
        raise NotImplementedError



import typing as t
from .launch_data_storage.session import SessionDataStorage

if t.TYPE_CHECKING:
    from .request import Request
    _JWT_BODY = t.Dict[str, object]
    _STATE_PARAMS = t.Dict[str, object]


class SessionService(object):
    data_storage = None  # type: SessionDataStorage[t.Any]
    _launch_data_lifetime = 86400  # type: int
    _session_prefix = 'lti1p3'  # type: str

    def __init__(self, request):
        # type: (Request) -> None
        self.data_storage = SessionDataStorage()
        self.data_storage.set_request(request)

    def _get_key(self, key, nonce=None, add_prefix=True):
        # type: (str, t.Optional[str], bool) -> str
        return ((self._session_prefix + '-') if add_prefix else '') + key + (('-' + nonce) if nonce else '')

    def _set_value(self, key, value):
        # type: (str, object) -> None
        self.data_storage.set_value(key, value, exp=self._launch_data_lifetime)

    def _get_value(self, key):
        # type: (str) -> t.Any
        return self.data_storage.get_value(key)

    def get_launch_data(self, key):
        # type: (str) -> _JWT_BODY
        return self._get_value(self._get_key(key, add_prefix=False))

    def save_launch_data(self, key, jwt_body):
        # type: (str, _JWT_BODY) -> None
        self._set_value(self._get_key(key, add_prefix=False), jwt_body)

    def save_nonce(self, nonce):
        # type: (str) -> None
        self._set_value(self._get_key('nonce', nonce), True)

    def check_nonce(self, nonce):
        # type: (str) -> bool
        nonce_key = self._get_key('nonce', nonce)
        return self.data_storage.check_value(nonce_key)

    def save_state_params(self, state, params):
        # type: (str, _STATE_PARAMS) -> None
        self._set_value(self._get_key(state), params)

    def get_state_params(self, state):
        # type: (str) -> _STATE_PARAMS
        return self._get_value(self._get_key(state))

    def set_state_valid(self, state, id_token_hash):
        # type: (str, str) -> None
        return self._set_value(self._get_key(state + '-id-token-hash'), id_token_hash)

    def check_state_is_valid(self, state, id_token_hash):
        # type: (str, str) -> bool
        return self._get_value(self._get_key(state + '-id-token-hash')) == id_token_hash

    def set_data_storage(self, data_storage):
        # type: (SessionDataStorage[t.Any]) -> None
        self.data_storage = data_storage

    def set_launch_data_lifetime(self, time_sec):
        # type: (int) -> None
        if self.data_storage.can_set_keys_expiration():
            self._launch_data_lifetime = time_sec
        else:
            raise Exception("%s launch storage doesn't support manual change expiration of the keys"
                            % self.data_storage.__class__.__name__)



from flask import make_response
from pylti1p3.oidc_login import OIDCLogin
from .cookie import FlaskCookieService
from .session import FlaskSessionService
from .redirect import FlaskRedirect


class FlaskOIDCLogin(OIDCLogin):

    def __init__(self, request, tool_config, session_service=None, cookie_service=None, launch_data_storage=None):
        cookie_service = cookie_service if cookie_service else FlaskCookieService(request)
        session_service = session_service if session_service else FlaskSessionService(request)
        super(FlaskOIDCLogin, self).__init__(request, tool_config, session_service, cookie_service, launch_data_storage)

    def get_redirect(self, url):
        return FlaskRedirect(url, self._cookie_service)

    def get_response(self, html):
        return make_response(html)
